#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import asyncio
import inspect
import itertools
import json
import os
import queue
import re
import shutil
import sys
import threading
import time
import warnings
from datetime import date, datetime, timezone
from typing import TYPE_CHECKING, Any, Callable, Dict, Literal, Optional, Tuple, Union
from urllib.parse import urljoin, urlparse

import humps
from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException, JobException
from hopsworks_common.constants import MODEL, PREDICTOR, Default
from hopsworks_common.core.constants import HAS_PANDAS
from hopsworks_common.git_file_status import GitFileStatus
from six import string_types


if HAS_PANDAS:
    import pandas as pd


FEATURE_STORE_NAME_SUFFIX = "_featurestore"


if TYPE_CHECKING:
    from hsfs import feature_group


class Encoder(json.JSONEncoder):
    def default(self, o: Any) -> Dict[str, Any]:
        try:
            return o.to_dict()
        except AttributeError:
            return super().default(o)


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types.
    Note that some numpy types doesn't have native python equivalence,
    hence json.dumps will raise TypeError.
    In this case, you'll need to convert your numpy types into its closest python equivalence.
    """

    def convert(self, obj):
        import base64

        import numpy as np

        def encode_binary(x):
            return base64.encodebytes(x).decode("ascii")

        if isinstance(obj, np.ndarray):
            if obj.dtype == np.object:
                return [self.convert(x)[0] for x in obj.tolist()]
            elif obj.dtype == np.bytes_:
                return np.vectorize(encode_binary)(obj), True
            else:
                return obj.tolist(), True

        if isinstance(obj, datetime) or (HAS_PANDAS and isinstance(obj, pd.Timestamp)):
            return obj.isoformat(), True
        if isinstance(obj, bytes) or isinstance(obj, bytearray):
            return encode_binary(obj), True
        if isinstance(obj, np.generic):
            return obj.item(), True
        if isinstance(obj, np.datetime64):
            return np.datetime_as_string(obj), True
        return obj, False

    def default(self, obj):  # pylint: disable=E0202
        res, converted = self.convert(obj)
        if converted:
            return res
        else:
            return super().default(obj)


VALID_EMBEDDING_TYPE = {
    "array<int>",
    "array<bigint>",
    "array<float>",
    "array<double>",
}


def validate_embedding_feature_type(embedding_index, schema):
    if not embedding_index or not schema:
        return
    feature_type_map = dict([(feat.name, feat.type) for feat in schema])
    for embedding in embedding_index.get_embeddings():
        feature_type = feature_type_map.get(embedding.name)
        if feature_type not in VALID_EMBEDDING_TYPE:
            raise FeatureStoreException(
                f"Provide feature `{embedding.name}` has type `{feature_type}`, "
                f"but requires one of the following: {', '.join(VALID_EMBEDDING_TYPE)}"
            )


def autofix_feature_name(name: str, warn: bool = False) -> str:
    # replace spaces with underscores and enforce lower case
    if warn and contains_uppercase(name):
        warnings.warn(
            "The feature name `{}` contains upper case letters. "
            "Feature names are sanitized to lower case in the feature store.".format(
                name
            ),
            stacklevel=1,
        )
    if warn and contains_whitespace(name):
        warnings.warn(
            "The feature name `{}` contains spaces. "
            "Feature names are sanitized to use underscore '_' in the feature store.".format(
                name
            ),
            stacklevel=1,
        )
    return name.lower().replace(" ", "_")


def contains_uppercase(name: str) -> bool:
    return any(re.finditer("[A-Z]", name))


def contains_whitespace(name: str) -> bool:
    return " " in name


def feature_group_name(
    feature_group,  #  FeatureGroup | ExternalFeatureGroup | SpineGroup
) -> str:
    return feature_group.name + "_" + str(feature_group.version)


def append_feature_store_suffix(name: str) -> str:
    name = name.lower()
    if name.endswith(FEATURE_STORE_NAME_SUFFIX):
        return name
    else:
        return name + FEATURE_STORE_NAME_SUFFIX


def strip_feature_store_suffix(name: str) -> str:
    name = name.lower()
    if name.endswith(FEATURE_STORE_NAME_SUFFIX):
        return name[: -1 * len(FEATURE_STORE_NAME_SUFFIX)]
    else:
        return name


def get_dataset_type(path: str) -> Literal["HIVEDB", "DATASET"]:
    if re.match(r"^(?:hdfs://|)/apps/hive/warehouse/*", path):
        return "HIVEDB"
    else:
        return "DATASET"


def check_timestamp_format_from_date_string(input_date: str) -> Tuple[str, str]:
    date_format_patterns = {
        r"^([0-9]{4})([0-9]{2})([0-9]{2})$": "%Y%m%d",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H%M",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})$": "%Y%m%d%H%M%S",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{3})$": "%Y%m%d%H%M%S%f",
        r"^([0-9]{4})([0-9]{2})([0-9]{2})T([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{6})Z$": "ISO",
    }
    normalized_date = (
        input_date.replace("/", "")
        .replace("-", "")
        .replace(" ", "")
        .replace(":", "")
        .replace(".", "")
    )

    date_format = None
    for pattern in date_format_patterns:
        date_format_pattern = re.match(pattern, normalized_date)
        if date_format_pattern:
            date_format = date_format_patterns[pattern]
            break

    if date_format is None:
        raise ValueError(
            "Unable to identify format of the provided date value : " + input_date
        )

    return normalized_date, date_format


def get_timestamp_from_date_string(input_date: str) -> int:
    norm_input_date, date_format = check_timestamp_format_from_date_string(input_date)
    try:
        if date_format != "ISO":
            date_time = datetime.strptime(norm_input_date, date_format)
        else:
            date_time = datetime.fromisoformat(input_date[:-1])
    except ValueError as err:
        raise ValueError(
            "Unable to parse the normalized input date value : "
            + norm_input_date
            + " with format "
            + date_format
        ) from err
    if date_time.tzinfo is None:
        date_time = date_time.replace(tzinfo=timezone.utc)
    return int(float(date_time.timestamp()) * 1000)


def get_hudi_datestr_from_timestamp(timestamp: int) -> str:
    return datetime.fromtimestamp(timestamp / 1000, timezone.utc).strftime(
        "%Y%m%d%H%M%S%f"
    )[:-3]


def get_delta_datestr_from_timestamp(timestamp: int) -> str:
    return datetime.fromtimestamp(timestamp / 1000, timezone.utc).strftime(
        "%Y-%m-%d %H:%M:%S.%f"
    )[:-3]


def convert_event_time_to_timestamp(
    event_time: Optional[
        Union[str, pd._libs.tslibs.timestamps.Timestamp, datetime, date, int]
    ],
) -> Optional[int]:
    if not event_time:
        return None
    if isinstance(event_time, str):
        return get_timestamp_from_date_string(event_time)
    elif hasattr(event_time, "to_pydatetime"):
        # only pandas Timestamp has to_pydatetime method out of the accepted event_time types
        # convert to unix epoch time in milliseconds.
        event_time = event_time.to_pydatetime()
        # convert to unix epoch time in milliseconds.
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, datetime):
        # convert to unix epoch time in milliseconds.
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, date):
        # convert to unix epoch time in milliseconds.
        event_time = datetime(*event_time.timetuple()[:7])
        if event_time.tzinfo is None:
            event_time = event_time.replace(tzinfo=timezone.utc)
        return int(event_time.timestamp() * 1000)
    elif isinstance(event_time, int):
        if event_time == 0:
            raise ValueError("Event time should be greater than 0.")
        # jdbc supports timestamp precision up to second only.
        if len(str(event_time)) <= 10:
            event_time = event_time * 1000
        return event_time
    else:
        raise ValueError(
            "Given event time should be in `datetime`, `date`, `str` or `int` type"
        )


def get_hostname_replaced_url(sub_path: str) -> str:
    """
    construct and return an url with public hopsworks hostname and sub path
    :param self:
    :param sub_path: url sub-path after base url
    :return: href url
    """
    href = urljoin(client.get_instance()._base_url, sub_path)
    url_parsed = client.get_instance().replace_public_host(urlparse(href))
    return url_parsed.geturl()


def verify_attribute_key_names(
    feature_group_obj,  #  FeatureGroup | ExternalFeatureGroup | SpineGroup
    external_feature_group: bool = False,
) -> None:
    feature_names = set(feat.name for feat in feature_group_obj.features)
    if feature_group_obj.primary_key:
        diff = set(feature_group_obj.primary_key) - feature_names
        if diff:
            raise FeatureStoreException(
                f"Provided primary key(s) {','.join(diff)} doesn't exist in feature dataframe"
            )

    if feature_group_obj.event_time:
        if feature_group_obj.event_time not in feature_names:
            raise FeatureStoreException(
                f"Provided event_time feature {feature_group_obj.event_time} doesn't exist in feature dataframe"
            )

    if not external_feature_group:
        if feature_group_obj.partition_key:
            diff = set(feature_group_obj.partition_key) - feature_names
            if diff:
                raise FeatureStoreException(
                    f"Provided partition key(s) {','.join(diff)} doesn't exist in feature dataframe"
                )

        if feature_group_obj.hudi_precombine_key:
            if feature_group_obj.hudi_precombine_key not in feature_names:
                raise FeatureStoreException(
                    f"Provided hudi precombine key {feature_group_obj.hudi_precombine_key} "
                    f"doesn't exist in feature dataframe"
                )


def get_job_url(href: str) -> str:
    """Use the endpoint returned by the API to construct the UI url for jobs

    Args:
        href (str): the endpoint returned by the API
    """
    url = urlparse(href)
    url_splits = url.path.split("/")
    project_id = url_splits[4]
    job_name = url_splits[6]
    ui_url = url._replace(
        path="p/{}/jobs/named/{}/executions".format(project_id, job_name)
    )
    ui_url = client.get_instance().replace_public_host(ui_url)
    return ui_url.geturl()


def _loading_animation(message: str, stop_event: threading.Event) -> None:
    for char in itertools.cycle([".", "..", "...", ""]):
        if stop_event.is_set():
            break
        print(f"{message}{char}   ", end="\r")
        time.sleep(0.5)


def run_with_loading_animation(message: str, func: Callable, *args, **kwargs) -> Any:
    stop_event = threading.Event()
    t = threading.Thread(
        target=_loading_animation,
        args=(
            message,
            stop_event,
        ),
    )
    t.daemon = True
    t.start()
    start = time.time()
    end = None

    try:
        result = func(*args, **kwargs)
        end = time.time()
        return result
    finally:
        # Stop the animation and print the "Finished Querying" message
        stop_event.set()
        t.join()
        if not end:
            print(f"\rError: {message}           ", end="\n")
        else:
            print(f"\rFinished: {message} ({(end-start):.2f}s) ", end="\n")


def get_feature_group_url(feature_store_id: int, feature_group_id: int) -> str:
    sub_path = (
        "/p/"
        + str(client.get_instance()._project_id)
        + "/fs/"
        + str(feature_store_id)
        + "/fg/"
        + str(feature_group_id)
    )
    return get_hostname_replaced_url(sub_path)


def is_runtime_notebook():
    if "ipykernel" in sys.modules:
        return True
    else:
        return False


class VersionWarning(Warning):
    pass


class ProvenanceWarning(Warning):
    pass


class JobWarning(Warning):
    pass


class StorageWarning(Warning):
    pass


class StatisticsWarning(Warning):
    pass


class ValidationWarning(Warning):
    pass


class FeatureGroupWarning(Warning):
    pass


def convert_to_abs(path, current_proj_name):
    abs_project_prefix = "/Projects/{}".format(current_proj_name)
    if not path.startswith(abs_project_prefix):
        return abs_project_prefix + "/" + path
    else:
        return path


def validate_job_conf(config, project_name):
    # User is required to set the appPath programmatically after getting the configuration
    if (
        config["type"] != "dockerJobConfiguration"
        and config["type"] != "flinkJobConfiguration"
        and "appPath" not in config
    ):
        raise JobException("'appPath' not set in job configuration")
    elif "appPath" in config and not config["appPath"].startswith("hdfs://"):
        config["appPath"] = "hdfs://" + convert_to_abs(config["appPath"], project_name)

    # If PYSPARK application set the mainClass, if SPARK validate there is a mainClass set
    if config["type"] == "sparkJobConfiguration":
        if config["appPath"].endswith(".py"):
            config["mainClass"] = "org.apache.spark.deploy.PythonRunner"
        elif "mainClass" not in config:
            raise JobException("'mainClass' not set in job configuration")

    return config


def convert_git_status_to_files(files):
    # Convert GitFileStatus to list of file paths
    if isinstance(files[0], GitFileStatus):
        tmp_files = []
        for file_status in files:
            tmp_files.append(file_status.file)
        files = tmp_files

    return files


def is_interactive():
    import __main__ as main

    return not hasattr(main, "__file__")


# Model registry

# - schema and types


def set_model_class(model):
    from hsml.llm.model import Model as LLMModel
    from hsml.model import Model as BaseModel
    from hsml.python.model import Model as PyModel
    from hsml.sklearn.model import Model as SkLearnModel
    from hsml.tensorflow.model import Model as TFModel
    from hsml.torch.model import Model as TorchModel

    if "href" in model:
        _ = model.pop("href")
    if "type" in model:  # backwards compatibility
        _ = model.pop("type")
    if "tags" in model:
        _ = model.pop("tags")  # tags are always retrieved from backend

    if "framework" not in model:
        return BaseModel(**model)

    framework = model.pop("framework")
    if framework == MODEL.FRAMEWORK_TENSORFLOW:
        return TFModel(**model)
    if framework == MODEL.FRAMEWORK_TORCH:
        return TorchModel(**model)
    if framework == MODEL.FRAMEWORK_SKLEARN:
        return SkLearnModel(**model)
    elif framework == MODEL.FRAMEWORK_PYTHON:
        return PyModel(**model)
    elif framework == MODEL.FRAMEWORK_LLM:
        return LLMModel(**model)
    else:
        raise ValueError(
            "framework {} is not a supported framework".format(str(framework))
        )


def input_example_to_json(input_example):
    import numpy as np

    if isinstance(input_example, np.ndarray):
        if input_example.size > 0:
            return _handle_tensor_input(input_example)
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_example))
            )
    elif isinstance(input_example, dict):
        return _handle_dict_input(input_example)
    else:
        return _handle_dataframe_input(input_example)


def _handle_tensor_input(input_tensor):
    return input_tensor.tolist()


def _handle_dataframe_input(input_ex):
    if HAS_PANDAS and isinstance(input_ex, pd.DataFrame):
        if not input_ex.empty:
            return input_ex.iloc[0].tolist()
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_ex))
            )
    elif HAS_PANDAS and isinstance(input_ex, pd.Series):
        if not input_ex.empty:
            return input_ex.tolist()
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_ex))
            )
    elif isinstance(input_ex, list):
        if len(input_ex) > 0:
            return input_ex
        else:
            raise ValueError(
                "input_example of type {} can not be empty".format(type(input_ex))
            )
    else:
        raise TypeError(
            "{} is not a supported input example type".format(type(input_ex))
        )


def _handle_dict_input(input_ex):
    return input_ex


# - artifacts


def compress(archive_out_path, archive_name, path_to_archive):
    if os.path.isdir(path_to_archive):
        return shutil.make_archive(
            os.path.join(archive_out_path, archive_name), "gztar", path_to_archive
        )
    else:
        return shutil.make_archive(
            os.path.join(archive_out_path, archive_name),
            "gztar",
            os.path.dirname(path_to_archive),
            os.path.basename(path_to_archive),
        )


def decompress(archive_file_path, extract_dir=None):
    return shutil.unpack_archive(archive_file_path, extract_dir=extract_dir)


# - export models


def validate_metrics(metrics):
    if metrics is not None:
        if not isinstance(metrics, dict):
            raise TypeError(
                "provided metrics is of instance {}, expected a dict".format(
                    type(metrics)
                )
            )

        for metric in metrics:
            # Validate key is a string
            if not isinstance(metric, string_types):
                raise TypeError(
                    "provided metrics key is of instance {}, expected a string".format(
                        type(metric)
                    )
                )
            # Validate value is a number
            try:
                float(metrics[metric])
            except ValueError as err:
                raise ValueError(
                    "{} is not a number, only numbers can be attached as metadata for models.".format(
                        str(metrics[metric])
                    )
                ) from err


# Model serving


def get_predictor_for_model(model, **kwargs):
    from hsml.llm.model import Model as LLMModel
    from hsml.llm.predictor import Predictor as vLLMPredictor
    from hsml.model import Model as BaseModel
    from hsml.predictor import Predictor as BasePredictor
    from hsml.python.model import Model as PyModel
    from hsml.python.predictor import Predictor as PyPredictor
    from hsml.sklearn.model import Model as SkLearnModel
    from hsml.sklearn.predictor import Predictor as SkLearnPredictor
    from hsml.tensorflow.model import Model as TFModel
    from hsml.tensorflow.predictor import Predictor as TFPredictor
    from hsml.torch.model import Model as TorchModel
    from hsml.torch.predictor import Predictor as TorchPredictor

    if not isinstance(model, BaseModel):
        raise ValueError(
            "model is of type {}, but an instance of {} class is expected".format(
                type(model), BaseModel
            )
        )

    if type(model) is TFModel:
        return TFPredictor(**kwargs)
    if type(model) is TorchModel:
        return TorchPredictor(**kwargs)
    if type(model) is SkLearnModel:
        return SkLearnPredictor(**kwargs)
    if type(model) is PyModel:
        return PyPredictor(**kwargs)
    if type(model) is LLMModel:
        return vLLMPredictor(**kwargs)
    if type(model) is BaseModel:
        return BasePredictor(  # python as default framework and model server
            model_framework=MODEL.FRAMEWORK_PYTHON,
            model_server=PREDICTOR.MODEL_SERVER_PYTHON,
            **kwargs,
        )


# General


def pretty_print(obj):
    if isinstance(obj, list):
        for logs in obj:
            pretty_print(logs)
    else:
        json_decamelized = humps.decamelize(obj.to_dict())
        print(json.dumps(json_decamelized, indent=4, sort_keys=True))


def get_members(cls, prefix=None):
    for m in inspect.getmembers(cls, lambda m: not (inspect.isroutine(m))):
        n = m[0]  # name
        if (prefix is not None and n.startswith(prefix)) or (
            prefix is None and not (n.startswith("__") and n.endswith("__"))
        ):
            yield m[1]  # value


# - json


def extract_field_from_json(obj, fields, default=None, as_instance_of=None):
    if isinstance(fields, list):
        for field in fields:
            value = extract_field_from_json(obj, field, default, as_instance_of)
            if value is not None:
                break
    else:
        value = obj.pop(fields) if fields in obj else default
        if as_instance_of is not None:
            if isinstance(value, list):
                # if the field is a list, get all obj
                value = [
                    get_obj_from_json(obj=subvalue, cls=as_instance_of)
                    for subvalue in value
                ]
            else:
                # otherwise, get single obj
                value = get_obj_from_json(obj=value, cls=as_instance_of)
    return value


def get_obj_from_json(obj, cls):
    if obj is not None:
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, dict):
            return cls.from_json(obj)
        if isinstance(obj, Default):
            return cls()
        raise ValueError(
            "Object of type {} cannot be converted to class {}".format(type(obj), cls)
        )
    return obj


def feature_view_to_json(obj):
    if obj is None:
        return None
    import importlib.util

    if importlib.util.find_spec("hsfs"):
        from hsfs import feature_view

        if isinstance(obj, feature_view.FeatureView):
            import json

            import humps

            return humps.camelize(json.loads(obj.json()))
    return None


def generate_fully_qualified_feature_name(
    feature_group: feature_group.FeatureGroup, feature_name: str
):
    """
    Generate the fully qualified feature name for a feature. The fully qualified name is created by concatenating
    the project name, feature group name, feature group version and feature name.
    """
    return f"{feature_group._get_project_name()}_{feature_group.name}_{feature_group.version}_{feature_name}"


class AsyncTask:
    """
    Generic class to represent an async task.

    Args:
        func (Callable): The function to run asynchronously.
        requires_connection_pool (bool): Whether the task requires a connection pool.
        **kwargs: Key word arguments to be passed to the functions.

    Properties:
        result (Any): The result of the async task.
        event (threading.Event): The event that will be set when the async task is finished.
    """

    def __init__(
        self,
        task_function: Callable,
        task_args: Tuple = (),
        requires_connection_pool=None,
        **kwargs,
    ):
        self.task_function = task_function
        self.task_args = task_args
        self.task_kwargs = kwargs
        self._event: threading.Event = threading.Event()
        self._result: Any = None
        self._requires_connection_pool = requires_connection_pool

    @property
    def result(self) -> Any:
        """
        The result of the async task.
        """
        return self._result

    @result.setter
    def result(self, value) -> None:
        self._result = value

    @property
    def event(self) -> threading.Event:
        """
        The event that will be set when the async task is finished.
        """
        return self._event

    @event.setter
    def event(self, value) -> None:
        self._event = value

    @property
    def requires_connection_pool(self) -> bool:
        """
        Whether the task requires a connection pool.
        """
        return self._requires_connection_pool


class AsyncTaskThread(threading.Thread):
    """
    Generic thread class that can be used to run async tasks in a separate thread.
    The thread will create its own event loop and run submitted tasks in that loop.

    The thread also store and fetches a connection pool that can be used by the async tasks.

    # Args:
        connection_pool_initializer (Callable): A function that initializes a connection pool.
        connection_pool_params (Tuple): The parameters to pass to the connection pool initializer.
        *thread_args: Arguments to be passed to the thread.
        **thread_kwargs: Key word arguments to be passed to the thread.

    # Properties:
        event_loop (asyncio.AbstractEventLoop): The event loop used by the thread.
        task_queue (queue.Queue[AsyncTask]): The queue used to submit tasks to the thread.
        connection_pool: The connection pool used
    """

    def __init__(
        self,
        connection_pool_initializer: Callable = None,
        connection_test: Callable = None,
        connection_pool_params: Tuple = (),
        *thread_args,
        **thread_kwargs,
    ):
        super().__init__(*thread_args, **thread_kwargs)
        self._task_queue: queue.Queue[AsyncTask] = queue.Queue()
        self._event_loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
        self.stop_event = threading.Event()
        self._connection_pool_initializer: Callable = connection_pool_initializer
        self._connection_test_function: Callable = connection_test
        self._connection_pool_params: Tuple = connection_pool_params
        self._connection_pool = None
        self.daemon = True  # Setting the thread as a daemon thread by default, so it will be terminated when the main thread is terminated.

    async def execute_task(self):
        """
        Execute the async tasks for the queue.
        """
        asyncio.set_event_loop(self._event_loop)

        while not self.stop_event.is_set():
            # Fetch a task from the queue.
            task = self.task_queue.get()
            # Run the task in the event loop and get the result
            try:
                if task.requires_connection_pool:
                    # Try checking connection to mysql and refresh it if required before running the task.
                    try:
                        await self._connection_test_function(self._connection_pool)
                    except Exception as e:
                        raise e

                    task.result = await task.task_function(
                        *task.task_args,
                        **task.task_kwargs,
                        connection_pool=self.connection_pool,
                    )
                else:
                    task.result = await task.task_function(
                        *task.task_args, **task.task_kwargs
                    )

                # Unblock the task, so the submit function can return the result.
                task.event.set()
            except Exception as e:
                task.result = e
                task.event.set()

    def stop(self):
        """
        Stop the thread and close the event loop.
        """
        self.stop_event.set()
        self._event_loop.stop()
        self._event_loop.close()

    def run(self):
        """
        Execute the async tasks for the queue.
        """
        asyncio.set_event_loop(self._event_loop)
        # Initialize the connection pool by using loop.run_until_complete to make sure the connection pool is initialized before the event loop starts running forever.
        if self._connection_pool_initializer:
            self._connection_pool = self._event_loop.run_until_complete(
                self._connection_pool_initializer(*self._connection_pool_params)
            )
        self._event_loop.create_task(self.execute_task())
        try:
            self._event_loop.run_forever()
        except Exception as e:
            print(
                "An error occurred in the async task thread the event loop has been closed: {}".format(
                    str(e)
                )
            )
            self._event_loop.stop()
            self._event_loop.close()
            # raise e
        finally:
            self._event_loop.close()

    def submit(self, task: AsyncTask):
        """
        Submit a async task to the thread and block until the execution of the function is completed.
        """
        # Submit a task to the queue.
        self.task_queue.put(task)
        # Block the execution until the task is finished.
        task.event.wait()

        if isinstance(task.result, Exception):
            raise task.result
        else:
            # Return the result of the task.
            return task.result

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        """
        The event loop used by the thread.
        """
        return self._event_loop

    @property
    def task_queue(self) -> queue.Queue[AsyncTask]:
        """
        The queue used to submit tasks to the thread.
        """
        return self._task_queue

    @property
    def connection_pool(self):
        """
        The connection pool used by the thread.
        """
        return self._connection_pool



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsml.connection import (
    HOPSWORKS_PORT_DEFAULT,
    HOSTNAME_VERIFICATION_DEFAULT,
    Connection,
)
from hsml.constants import HOSTS


class TestConnection:
    # constants

    def test_constants(self):
        # The purpose of this test is to ensure that (1) we don't make undesired changes to contant values
        # that might break things somewhere else, and (2) we remember to update the pytests accordingly by
        # adding / removing / updating tests, if necessary.
        assert HOSTS.APP_HOST == "c.app.hopsworks.ai"
        assert HOPSWORKS_PORT_DEFAULT == 443
        assert HOSTNAME_VERIFICATION_DEFAULT is False

    # constructor

    def test_constructor_default(self, mocker):
        # Arrange
        class MockConnection:
            pass

        mock_connection = MockConnection()
        mock_connection.connect = mocker.MagicMock()
        mock_connection.init = Connection.__init__

        # Act
        mock_connection.init(mock_connection)

        # Assert
        assert mock_connection._host is None
        assert mock_connection._port == HOPSWORKS_PORT_DEFAULT
        assert mock_connection._project is None
        assert mock_connection._hostname_verification == HOSTNAME_VERIFICATION_DEFAULT
        assert mock_connection._trust_store_path is None
        assert mock_connection._api_key_file is None
        assert mock_connection._api_key_value is None
        assert not mock_connection._connected
        mock_connection.connect.assert_called_once()

    def test_constructor(self, mocker):
        # Arrange
        class MockConnection:
            pass

        mock_connection = MockConnection()
        mock_connection.connect = mocker.MagicMock()
        mock_connection.init = Connection.__init__

        # Act
        mock_connection.init(
            mock_connection,
            host="host",
            port=1234,
            project="project",
            hostname_verification=False,
            trust_store_path="ts_path",
            api_key_file="ak_file",
            api_key_value="ak_value",
        )

        # Assert
        assert mock_connection._host == "host"
        assert mock_connection._port == 1234
        assert mock_connection._project == "project"
        assert not mock_connection._hostname_verification
        assert mock_connection._trust_store_path == "ts_path"
        assert mock_connection._api_key_file == "ak_file"
        assert mock_connection._api_key_value == "ak_value"
        assert not mock_connection._connected
        mock_connection.connect.assert_called_once()

    # handlers

    def test_get_model_registry(self, mocker):
        # Arrange
        mock_connection = mocker.MagicMock()
        mock_connection.get_model_registry = Connection.get_model_registry
        mock_connection._model_registry_api = mocker.MagicMock()
        mock_connection._model_registry_api.get = mocker.MagicMock(return_value="mr")

        # Act
        mr = mock_connection.get_model_registry(mock_connection)

        # Assert
        assert mr == "mr"
        mock_connection._model_registry_api.get.assert_called_once()

    def test_get_model_serving(self, mocker):
        # Arrange
        mock_connection = mocker.MagicMock()
        mock_connection.get_model_serving = Connection.get_model_serving
        mock_connection._model_serving_api = mocker.MagicMock()
        mock_connection._model_serving_api.get = mocker.MagicMock(return_value="ms")

        # Act
        ms = mock_connection.get_model_serving(mock_connection)

        # Assert
        assert ms == "ms"
        mock_connection._model_serving_api.get.assert_called_once()

    # connection

    # TODO: Add tests for connection-related methods

    def test_connect(self, mocker):
        pass

    def test_close(self, mocker):
        pass

    def test_connection(self, mocker):
        pass



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import ast
import copy
import inspect
import json
import re
import warnings
from dataclasses import dataclass
from datetime import date, datetime, time
from enum import Enum
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import humps
from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException
from hopsworks_common.constants import FEATURES
from hsfs import engine, util
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics
from hsfs.decorators import typechecked
from hsfs.transformation_statistics import TransformationStatistics
from packaging.version import Version


class UDFExecutionMode(Enum):
    """
    Class that store the possible execution types of UDF's.
    """

    DEFAULT = "default"
    PYTHON = "python"
    PANDAS = "pandas"

    def get_current_execution_mode(self, online):
        if self == UDFExecutionMode.DEFAULT and online:
            return UDFExecutionMode.PYTHON
        elif self == UDFExecutionMode.DEFAULT and not online:
            return UDFExecutionMode.PANDAS
        else:
            return self

    @staticmethod
    def from_string(execution_mode: str):
        try:
            return UDFExecutionMode[execution_mode.upper()]
        except KeyError as e:
            raise FeatureStoreException(
                f"Ivalid execution mode `{execution_mode}` for UDF. Please use `default`, `python` or `pandas` instead."
            ) from e


class UDFKeyWords(Enum):
    """
    Class that stores the keywords used as arguments in a UDFs.
    """

    STATISTICS = "statistics"
    CONTEXT = "context"


def udf(
    return_type: Union[List[type], type],
    drop: Optional[Union[str, List[str]]] = None,
    mode: Literal["default", "python", "pandas"] = "default",
) -> "HopsworksUdf":
    """
    Create an User Defined Function that can be and used within the Hopsworks Feature Store to create transformation functions.

    Hopsworks UDF's are user defined functions that executes as 'pandas_udf' when executing
    in spark engine and as pandas functions in the python engine. The pandas udf/pandas functions
    gets as inputs pandas Series's and can provide as output a pandas Series or a pandas DataFrame.
    A Hopsworks udf is defined using the `hopsworks_udf` decorator. The outputs of the defined UDF
    must be mentioned in the decorator as a list of python types.


    !!! example
        ```python
        from hopsworks import udf

        @udf(float)
        def add_one(data1):
            return data1 + 1
        ```

    # Arguments
        return_type: `Union[List[type], type]`. The output types of the defined UDF
        drop: `Optional[Union[str, List[str]]]`. The features to be dropped after application of transformation functions. Default's to None.
        mode: `Literal["default", "python", "pandas"]`. The exection mode of the UDF. Default's to 'default'

    # Returns
        `HopsworksUdf`: The metadata object for hopsworks UDF's.

    # Raises
        `hopsworks.client.exceptions.FeatureStoreException` : If unable to create UDF.
    """

    def wrapper(func: Callable) -> HopsworksUdf:
        udf = HopsworksUdf(
            func=func,
            return_types=return_type,
            dropped_argument_names=drop,
            execution_mode=UDFExecutionMode.from_string(mode),
        )
        return udf

    return wrapper


@dataclass
class TransformationFeature:
    """
    Mapping of feature names to their corresponding statistics argument names in the code.

    The statistic_argument_name for a feature name would be None if the feature does not need statistics.

    # Arguments
        feature_name : `str`. Name of the feature.
        statistic_argument_name : `str`. Name of the statistics argument in the code for the feature specified in the feature name.
    """

    feature_name: str
    statistic_argument_name: Optional[str]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "feature_name": self.feature_name,
            "statistic_argument_name": self.statistic_argument_name,
        }


@typechecked
class HopsworksUdf:
    """
    Meta data for user defined functions.

    Stores meta data required to execute the user defined function in both spark and python engine.
    The class generates uses the metadata to dynamically generate user defined functions based on the
    engine it is executed in.

    # Arguments
        func : `Union[Callable, str]`. The transformation function object or the source code of the transformation function.
        return_types : `Union[List[type], type, List[str], str]`. A python type or a list of python types that denotes the data types of the columns output from the transformation functions.
        name : `Optional[str]`. Name of the transformation function.
        transformation_features : `Optional[List[TransformationFeature]]`. A list of objects of `TransformationFeature` that maps the feature used for transformation to their corresponding statistics argument names if any
        transformation_function_argument_names : `Optional[List[str]]`. The argument names of the transformation function.
        dropped_argument_names : `Optional[List[str]]`. The arguments to be dropped from the finial DataFrame after the transformation functions are applied.
        dropped_feature_names : `Optional[List[str]]`. The feature name corresponding to the arguments names that are dropped
        feature_name_prefix: `Optional[str]`. Prefixes if any used in the feature view.
        output_column_names: `Optional[List[str]]`. The names of the output columns returned from the transformation function.
        generate_output_col_names: `bool`. Generate default output column names for the transformation function. Default's to True.
    """

    # Mapping for converting python types to spark types - required for creating pandas UDF's.
    PYTHON_SPARK_TYPE_MAPPING = {
        str: "string",
        int: "bigint",
        float: "double",
        bool: "boolean",
        datetime: "timestamp",
        time: "timestamp",
        date: "date",
    }

    def __init__(
        self,
        func: Union[Callable, str],
        return_types: Union[List[type], type, List[str], str],
        execution_mode: UDFExecutionMode,
        name: Optional[str] = None,
        transformation_features: Optional[List[TransformationFeature]] = None,
        transformation_function_argument_names: Optional[List[str]] = None,
        dropped_argument_names: Optional[List[str]] = None,
        dropped_feature_names: Optional[List[str]] = None,
        feature_name_prefix: Optional[str] = None,
        output_column_names: Optional[str] = None,
        generate_output_col_names: bool = True,
    ):
        self._return_types: List[str] = HopsworksUdf._validate_and_convert_output_types(
            return_types
        )

        self._execution_mode: UDFExecutionMode = execution_mode

        self._feature_name_prefix: Optional[str] = (
            feature_name_prefix  # Prefix to be added to feature names
        )

        self._function_name: str = func.__name__ if name is None else name

        self._function_source: str = (
            HopsworksUdf._extract_source_code(func)
            if isinstance(func, Callable)
            else func
        )

        # The parameter `output_column_names` is initialized lazily.
        # It is only initialized if the output column names are retrieved from the backend or explicitly specified using the `alias` function or is initialized with default column names if the UDF is accessed from a transformation function.
        # Output column names are only stored in the backend when a model dependent or on demand transformation function is created using the defined UDF.
        self._output_column_names: List[str] = []

        if not transformation_features:
            # New transformation function being declared so extract source code from function
            self._transformation_features: List[TransformationFeature] = (
                HopsworksUdf._extract_function_arguments(func)
                if not transformation_features
                else transformation_features
            )

            self._transformation_function_argument_names = [
                feature.feature_name for feature in self._transformation_features
            ]

            self._dropped_argument_names: List[str] = (
                HopsworksUdf._validate_and_convert_drop_features(
                    dropped_argument_names,
                    self.transformation_features,
                    feature_name_prefix,
                )
            )
            self._dropped_features = self._dropped_argument_names

        else:
            self._transformation_features = transformation_features
            self._transformation_function_argument_names = (
                transformation_function_argument_names
            )
            self._dropped_argument_names = dropped_argument_names
            self._dropped_features = (
                dropped_feature_names
                if dropped_feature_names
                else dropped_argument_names
            )
            self._output_column_names = (
                output_column_names if output_column_names else []
            )

        self._formatted_function_source, self._module_imports = (
            HopsworksUdf._format_source_code(self._function_source)
        )

        self._statistics: Optional[TransformationStatistics] = None

        self._transformation_context: Dict[str, Any] = {}

        # Denote if the output feature names have to be generated.
        # Set to `False` if the output column names are saved in the backend and the udf is constructed from it using `from_response_json` function or if user has specified the output feature names using the `alias`` function.
        self._generate_output_col_name: bool = generate_output_col_names

    @staticmethod
    def _validate_and_convert_drop_features(
        dropped_features: Union[str, List[str]],
        transformation_feature: List[str],
        feature_name_prefix: str,
    ) -> List[str]:
        """
        Function that converts dropped features to a list and validates if the dropped feature is present in the transformation function
        # Arguments
            dropped_features: `Union[str, List[str]]`. Features of be dropped.
            transformation_feature: `List[str]`. Features to be transformed in the UDF
        # Returns
            `List[str]`: A list of features to be dropped.
        """
        if not dropped_features:
            return None

        dropped_features = (
            [dropped_features]
            if not isinstance(dropped_features, list)
            else dropped_features
        )

        feature_name_prefix = feature_name_prefix if feature_name_prefix else ""

        missing_drop_features = []
        for dropped_feature in dropped_features:
            dropped_feature = feature_name_prefix + dropped_feature
            if dropped_feature not in transformation_feature:
                missing_drop_features.append(dropped_feature)

        if missing_drop_features:
            missing_drop_features = "', '".join(missing_drop_features)
            raise FeatureStoreException(
                f"Cannot drop features '{missing_drop_features}' as they are not features given as arguments in the defined UDF."
            )

        return dropped_features

    @staticmethod
    def _validate_and_convert_output_types(
        output_types: Union[List[type], List[str]],
    ) -> List[str]:
        """
        Function that takes in a type or list of types validates if it is supported and return a list of strings

        # Arguments
            output_types: `list`. List of python types.

        # Raises
            `hopsworks.client.exceptions.FeatureStoreException` : If any of the output type is invalid
        """
        convert_output_types = []
        output_types = (
            output_types if isinstance(output_types, List) else [output_types]
        )
        for output_type in output_types:
            if (
                output_type not in HopsworksUdf.PYTHON_SPARK_TYPE_MAPPING.keys()
                and output_type not in HopsworksUdf.PYTHON_SPARK_TYPE_MAPPING.values()
            ):
                raise FeatureStoreException(
                    f"Output type {output_type} is not supported. Please refer to the documentation to get more information on the supported types."
                )
            convert_output_types.append(
                output_type
                if isinstance(output_type, str)
                else HopsworksUdf.PYTHON_SPARK_TYPE_MAPPING[output_type]
            )
        return convert_output_types

    @staticmethod
    def _get_module_imports(path: str) -> List[str]:
        """Function that extracts the imports used in the python file specified in the path.

        # Arguments
            path: `str`. Path to python file from which imports are to be extracted.

        # Returns
            `List[str]`: A list of string that contains the import statement using in the file.
        """
        imports = []
        with open(path) as fh:
            root = ast.parse(fh.read(), path)
        for node in ast.iter_child_nodes(root):
            if isinstance(node, ast.Import):
                imported_module = False
            elif isinstance(node, ast.ImportFrom):
                imported_module = node.module
            else:
                continue
            for n in node.names:
                if imported_module:
                    import_line = "from " + imported_module + " import " + n.name
                elif n.asname:
                    import_line = "import " + n.name + " as " + n.asname
                else:
                    import_line = "import " + n.name
                imports.append(import_line)
        return imports

    @staticmethod
    def _extract_source_code(udf_function: Callable) -> str:
        """
        Function to extract the source code of the function along with the imports used in the file.

        The module imports cannot be extracted if the function is defined in a jupyter notebook.

        # Arguments
            udf_function: `Callable`. Function for which the source code must be extracted.
        # Returns
            `str`: a string that contains the source code of function along with the extracted module imports.
        """
        try:
            module_imports = HopsworksUdf._get_module_imports(
                inspect.getfile(udf_function)
            )
        except FileNotFoundError:
            module_imports = [""]
            warnings.warn(
                "Cannot extract imported dependencies for the UDF from the module in which it is defined. Please make sure to import all dependencies for the UDF inside the function.",
                stacklevel=2,
            )

        function_code = inspect.getsource(udf_function)
        source_code = "\n".join(module_imports) + "\n" + function_code

        return source_code

    @staticmethod
    def _parse_function_signature(source_code: str) -> Tuple[List[str], str, int, int]:
        """
        Function to parse the source code to extract the argument along with the start and end line of the function signature

        # Arguments
            source_code: `str`. Source code of a function.
        # Returns
            `List[str]`: List of function arguments
            `str`: function signature
            `int`: starting line number of function signature
            `int`: ending line number of function signature

        """
        source_code = source_code.split("\n")

        signature_start_line = None
        signature_end_line = None
        # Find the line where the function signature is defined
        for i, line in enumerate(source_code):
            if line.strip().startswith("def "):
                signature_start_line = i
            if signature_start_line is not None and re.search(
                r"\)\s*(->.*)?\s*:$", line.split("#")[0].strip()
            ):
                signature_end_line = i
                break

        # Parse the function signature to remove the specified argument
        signature = "".join(
            [
                code.split("#")[0]
                for code in source_code[signature_start_line : signature_end_line + 1]
            ]
        )
        arg_list = signature.split("(")[1].split(")")[0].split(",")
        arg_list = [
            arg.split(":")[0].split("=")[0].strip()
            for arg in arg_list
            if not arg.strip() == ""
        ]

        # Extracting keywords like `context`and `statistics` from the function signature.
        keyword_to_remove = [keyword.value for keyword in UDFKeyWords]
        arg_list = [arg for arg in arg_list if arg not in keyword_to_remove]

        return arg_list, signature, signature_start_line, signature_end_line

    @staticmethod
    def _extract_function_arguments(function: Callable) -> List[TransformationFeature]:
        """
        Function to extract the argument names from a provided function source code.

        # Arguments
            source_code: `Callable`. The function for which the value are to be extracted.
        # Returns
            `List[TransformationFeature]`: List of TransformationFeature that provide a mapping from feature names to corresponding statistics parameters if any is present.
        """
        arg_list = []
        statistics = None
        signature = inspect.signature(function).parameters
        if not signature:
            raise FeatureStoreException(
                "No arguments present in the provided user defined function. Please provide at least one argument in the defined user defined function."
            )

        # Extracting keywords like `context`and `statistics` from the function signature.
        for arg in inspect.signature(function).parameters.values():
            # Fetching default value for statistics parameter to find argument names that require statistics.
            if arg.name == UDFKeyWords.STATISTICS.value:
                statistics = arg.default
            elif arg.name != UDFKeyWords.CONTEXT.value:
                arg_list.append(arg.name)

        if statistics:
            missing_statistic_features = [
                statistic_feature
                for statistic_feature in statistics._features
                if statistic_feature not in arg_list
            ]
            if missing_statistic_features:
                missing_statistic_features = "', '".join(missing_statistic_features)
                raise FeatureStoreException(
                    f"No argument corresponding to statistics parameter '{missing_statistic_features}' present in function definition."
                )
            return [
                TransformationFeature(arg, arg if arg in statistics._features else None)
                for arg in arg_list
            ]
        else:
            return [TransformationFeature(arg, None) for arg in arg_list]

    @staticmethod
    def _format_source_code(source_code: str) -> Tuple[str, str]:
        """
        Function that parses the existing source code to remove statistics parameter and remove all decorators and type hints from the function source code.

        # Arguments
            source_code: `str`. Source code of a function.
        # Returns
            `Tuple[str, str]`: Tuple that contains Source code that does not contain any decorators, type hints or statistics parameters and the module imports
        """

        arg_list, signature, _, signature_end_line = (
            HopsworksUdf._parse_function_signature(source_code)
        )
        module_imports = source_code.split("@")[0]

        # Reconstruct the function signature
        new_signature = (
            signature.split("(")[0].strip() + "(" + ", ".join(arg_list) + "):"
        )
        source_code = source_code.split("\n")
        # Reconstruct the modified function as a string
        modified_source = (
            new_signature + "\n\t" + "\n\t".join(source_code[signature_end_line + 1 :])
        )

        return modified_source, module_imports

    def _create_pandas_udf_return_schema_from_list(self) -> str:
        """
        Function that creates the return schema required for executing the defined UDF's as pandas UDF's in Spark.

        # Returns
            `str`: DDL-formatted type string that denotes the return types of the user defined function.
        """
        if len(self.return_types) > 1:
            return ", ".join(
                [
                    f"`{self.output_column_names[i]}` {self.return_types[i]}"
                    for i in range(len(self.return_types))
                ]
            )
        else:
            return self.return_types[0]

    def _prepare_transformation_function_scope(self, **kwargs) -> Dict[str, Any]:
        """
        Function that prepares the scope for the transformation function to be executed. This scope would include any variable that are required to be injected into the transformation function.

        By default the output column names, transformation statistics and transformation context are injected into the scope if they are required by the transformation function. Any additional arguments can be passed as kwargs.
        """
        # Shallow copy of scope performed because updating statistics argument of scope must not affect other instances.
        scope = __import__("__main__").__dict__.copy()

        # Adding variables required to be injected into the scope.
        vaariable_to_inject = {
            UDFKeyWords.STATISTICS.value: self.transformation_statistics,
            UDFKeyWords.CONTEXT.value: self.transformation_context,
            "_output_col_names": self.output_column_names,
        }
        vaariable_to_inject.update(**kwargs)

        # Injecting variables that have a value into scope.
        scope.update({k: v for k, v in vaariable_to_inject.items() if v is not None})

        return scope

    def python_udf_wrapper(self, rename_outputs) -> Callable:
        """
        Function that creates a dynamic wrapper function for the defined udf. The wrapper function would be used to specify column names, in spark engines and to localize timezones.

        The renames is done so that the column names match the schema expected by spark when multiple columns are returned in a spark udf.
        The wrapper function would be available in the main scope of the program.

        # Returns
            `Callable`: A wrapper function that renames outputs of the User defined function into specified output column names.
        """
        # Check if any output is of date time type.
        date_time_output_index = [
            ind for ind, ele in enumerate(self.return_types) if ele == "timestamp"
        ]

        # Function that converts the timestamp to localized timezone
        convert_timstamp_function = (
            "def convert_timezone(date_time_obj : datetime):\n"
            + "   from datetime import datetime, timezone\n"
            + "   import tzlocal\n"
            + "   current_timezone = tzlocal.get_localzone()\n"
            + "   if date_time_obj and isinstance(date_time_obj, datetime):\n"
            + "      if date_time_obj.tzinfo is None:\n"
            + "      # if timestamp is timezone unaware, make sure it's localized to the system's timezone.\n"
            + "      # otherwise, spark will implicitly convert it to the system's timezone.\n"
            + "         return date_time_obj.replace(tzinfo=current_timezone)\n"
            + "      else:\n"
            + "         return date_time_obj.astimezone(timezone.utc).replace(tzinfo=current_timezone)\n"
            + "   else:\n"
            + "      return None\n"
        )

        # Start wrapper function generation
        code = (
            self._module_imports
            + "\n"
            + (convert_timstamp_function + "\n" if date_time_output_index else "\n")
            + "def wrapper(*args):\n"
            + f"   {self._formatted_function_source}\n"
            + f"   transformed_features = {self.function_name}(*args)\n"
        )
        if len(self.return_types) > 1:
            # If date time columns are there convert make sure that they are localized.
            if date_time_output_index:
                code += (
                    "   transformed_features = list(transformed_features)\n"
                    "   for index in _date_time_output_index:\n"
                    + "      transformed_features[index] = convert_timezone(transformed_features[index])\n"
                )
            if rename_outputs:
                # Use a dictionary to rename output to correct column names. This must be for the udf's to be executable in spark.
                code += "   return dict(zip(_output_col_names, transformed_features))"
            else:
                code += "   return transformed_features"
        else:
            if date_time_output_index:
                code += (
                    "   transformed_features = convert_timezone(transformed_features)\n"
                )
            code += "   return transformed_features"

        # Inject required parameter to scope
        scope = self._prepare_transformation_function_scope(
            _date_time_output_index=date_time_output_index
        )

        # executing code
        exec(code, scope)

        # returning executed function object
        return eval("wrapper", scope)

    def pandas_udf_wrapper(self) -> Callable:
        """
        Function that creates a dynamic wrapper function for the defined udf that renames the columns output by the UDF into specified column names.

        The renames is done so that the column names match the schema expected by spark when multiple columns are returned in a pandas udf.
        The wrapper function would be available in the main scope of the program.

        # Returns
            `Callable`: A wrapper function that renames outputs of the User defined function into specified output column names.
        """

        date_time_output_columns = [
            self.output_column_names[ind]
            for ind, ele in enumerate(self.return_types)
            if ele == "timestamp"
        ]

        # Function to make transformation function time safe. Defined as a string because it has to be dynamically injected into scope to be executed by spark
        convert_timstamp_function = """def convert_timezone(date_time_col : pd.Series):
        import tzlocal
        current_timezone = tzlocal.get_localzone()
        if date_time_col.dt.tz is None:
            # if timestamp is timezone unaware, make sure it's localized to the system's timezone.
            # otherwise, spark will implicitly convert it to the system's timezone.
            return date_time_col.dt.tz_localize(str(current_timezone))
        else:
            # convert to utc, then localize to system's timezone
            return date_time_col.dt.tz_convert('UTC').dt.tz_localize(None).dt.tz_localize(str(current_timezone))"""

        # Defining wrapper function that renames the column names to specific names
        if len(self.return_types) > 1:
            code = (
                self._module_imports
                + "\n"
                + f"""import pandas as pd
{convert_timstamp_function}
def renaming_wrapper(*args):
    {self._formatted_function_source}
    df = {self.function_name}(*args)
    if isinstance(df, tuple):
        df = pd.concat(df, axis=1)
    df.columns = _output_col_names
    for col in _date_time_output_columns:
        if pd.api.types.is_datetime64_any_dtype(df[col]):
            df[col] = convert_timezone(df[col])
    return df"""
            )
        else:
            code = (
                self._module_imports
                + "\n"
                + f"""import pandas as pd
{convert_timstamp_function}
def renaming_wrapper(*args):
    {self._formatted_function_source}
    df = {self.function_name}(*args)
    df = df.rename(_output_col_names[0])
    if _date_time_output_columns:
        # Set correct type is column is not of datetime type
        if pd.api.types.is_datetime64_any_dtype(df):
            df = convert_timezone(df)
    return df"""
            )

        # Inject required parameter to scope
        scope = self._prepare_transformation_function_scope(
            _date_time_output_columns=date_time_output_columns
        )

        # executing code
        exec(code, scope)

        # returning executed function object
        return eval("renaming_wrapper", scope)

    def __call__(self, *features: List[str]) -> "HopsworksUdf":
        """
        Set features to be passed as arguments to the user defined functions

        # Arguments
            features: Name of features to be passed to the User Defined function
        # Returns
            `HopsworksUdf`: Meta data class for the user defined function.
        # Raises
            `hopsworks.client.exceptions.FeatureStoreException`: If the provided number of features do not match the number of arguments in the defined UDF or if the provided feature names are not strings.
        """

        if len(features) != len(self.transformation_features):
            raise FeatureStoreException(
                "Number of features provided does not match the number of features provided in the UDF definition"
            )

        for arg in features:
            if not isinstance(arg, str):
                raise FeatureStoreException(
                    f'Feature names provided must be string "{arg}" is not string'
                )
        transformation_feature_name = self.transformation_features
        if self.dropped_features:
            index_dropped_features = [
                transformation_feature_name.index(dropped_feature)
                for dropped_feature in self.dropped_features
            ]
            updated_dropped_features = [
                features[index] for index in index_dropped_features
            ]
        else:
            updated_dropped_features = None

        # Create a copy of the UDF to associate it with new feature names.
        udf = copy.deepcopy(self)

        udf._transformation_features = [
            TransformationFeature(
                new_feature_name, transformation_feature.statistic_argument_name
            )
            for transformation_feature, new_feature_name in zip(
                self._transformation_features, features
            )
        ]
        udf.dropped_features = updated_dropped_features
        return udf

    def alias(self, *args: str):
        """
        Set the names of the transformed features output by the UDF.
        """
        if len(args) == 1 and isinstance(args[0], list):
            # If a single list is passed, use it directly
            output_col_names = args[0]
        else:
            # Otherwise, use the individual arguments as a list
            output_col_names = list(args)
        if any(
            not isinstance(output_col_name, str) for output_col_name in output_col_names
        ):
            raise FeatureStoreException(
                f"Invalid output feature names provided for the transformation function '{repr(self)}'. Please ensure all arguments are strings."
            )

        self._generate_output_col_name = False
        self.output_column_names = output_col_names

        return self

    def _validate_output_col_name(self, output_col_names):
        if any(
            len(output_col_name) > FEATURES.MAX_LENGTH_NAME
            for output_col_name in output_col_names
        ):
            raise FeatureStoreException(
                f"Invalid output feature names specified for the transformation function '{repr(self)}'. Please provide names shorter than {FEATURES.MAX_LENGTH_NAME} characters."
            )

        if len(output_col_names) != len(set(output_col_names)):
            raise FeatureStoreException(
                f"Duplicate output feature names provided for the transformation function '{repr(self)}'. Please ensure all arguments names are unique."
            )

        if output_col_names and len(output_col_names) != len(self.return_types):
            raise FeatureStoreException(
                f"The number of output feature names provided does not match the number of features returned by the transformation function '{repr(self)}'. Pease provide exactly {len(self.return_types)} feature name(s) to match the output."
            )

    def _validate_transformation_context(self, transformation_context: Dict[str, Any]):
        """
        Function that checks if the context variables provided to the transformation function is valid.

        it checks if context variables are defined as a dictionary.
        """
        if not isinstance(transformation_context, dict):
            raise FeatureStoreException(
                "Transformation context variable must be passed as dictionary."
            )

        return transformation_context

    def update_return_type_one_hot(self):
        self._return_types = [
            self._return_types[0]
            for _ in range(len(self.transformation_statistics.feature.unique_values))
        ]

    def get_udf(self, online: bool = False) -> Callable:
        """
        Function that checks the current engine type, execution type and returns the appropriate UDF.

        If the execution mode is : "default":
            - In the `spark` engine : During inference a spark udf is returned otherwise a spark pandas_udf is returned.
            - In the `python` engine : During inference a python udf is returned otherwise a pandas udf is returned.
        If the execution mode is : "pandas":
            - In the `spark` engine : Always returns a spark pandas udf.
            - In the `python` engine : Always returns a pandas udf.
        If the execution mode is : "python":
            - In the `spark` engine : Always returns a spark udf.
            - In the `python` engine : Always returns a python udf.

        # Arguments
            inference: `bool`. Specify if udf required for online inference.

        # Returns
            `Callable`: Pandas UDF in the spark engine otherwise returns a python function for the UDF.
        """

        if (
            self.execution_mode.get_current_execution_mode(online)
            == UDFExecutionMode.PANDAS
        ):
            if engine.get_type() in ["python", "training"] or online:
                return self.pandas_udf_wrapper()
            else:
                from pyspark.sql.functions import pandas_udf

                return pandas_udf(
                    f=self.pandas_udf_wrapper(),
                    returnType=self._create_pandas_udf_return_schema_from_list(),
                )
        elif (
            self.execution_mode.get_current_execution_mode(online)
            == UDFExecutionMode.PYTHON
        ):
            if engine.get_type() in ["python", "training"] or online:
                # Renaming into correct column names done within Python engine since a wrapper does not work for polars dataFrames.
                return self.python_udf_wrapper(rename_outputs=False)
            else:
                from pyspark.sql.functions import udf as pyspark_udf

                return pyspark_udf(
                    f=self.python_udf_wrapper(rename_outputs=True),
                    returnType=self._create_pandas_udf_return_schema_from_list(),
                )

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert class into a dictionary.

        # Returns
            `Dict`: Dictionary that contains all data required to json serialize the object.
        """
        backend_version = client.get_connection().backend_version

        return {
            "sourceCode": self._function_source,
            "outputTypes": self.return_types,
            "transformationFeatures": self.transformation_features,
            "transformationFunctionArgumentNames": self._transformation_function_argument_names,
            "droppedArgumentNames": self._dropped_argument_names,
            "statisticsArgumentNames": self._statistics_argument_names
            if self.statistics_required
            else None,
            "name": self.function_name,
            "featureNamePrefix": self._feature_name_prefix,
            "executionMode": self.execution_mode.value.upper(),
            **(
                {"outputColumnNames": self.output_column_names}
                if Version(backend_version) >= Version("4.1.6")
                else {}
            ),  # This check is added for backward compatibility with older versions of Hopsworks. The "outputColumnNames" field was added in Hopsworks 4.1.6 and versions below do not support unknown fields in the backend.
        }

    def json(self) -> str:
        """
        Convert class into its json serialized form.

        # Returns
            `str`: Json serialized object.
        """
        return json.dumps(self, cls=util.Encoder)

    @classmethod
    def from_response_json(
        cls: "HopsworksUdf", json_dict: Dict[str, Any]
    ) -> "HopsworksUdf":
        """
        Function that constructs the class object from its json serialization.

        # Arguments
            json_dict: `Dict[str, Any]`. Json serialized dictionary for the class.
        # Returns
            `HopsworksUdf`: Json deserialized class object.
        """

        json_decamelized = humps.decamelize(json_dict)
        function_source_code = json_decamelized["source_code"]
        function_name = json_decamelized["name"]
        feature_name_prefix = json_decamelized.get("feature_name_prefix", None)
        output_types = [
            output_type.strip() for output_type in json_decamelized["output_types"]
        ]
        transformation_features = [
            feature.strip() for feature in json_decamelized["transformation_features"]
        ]
        dropped_argument_names = (
            [
                dropped_feature.strip()
                for dropped_feature in json_decamelized["dropped_argument_names"]
            ]
            if json_decamelized.get("dropped_argument_names", None)
            else None
        )
        transformation_function_argument_names = (
            [
                arg_name.strip()
                for arg_name in json_decamelized[
                    "transformation_function_argument_names"
                ]
            ]
            if json_decamelized.get("transformation_function_argument_names", None)
            else None
        )
        statistics_features = (
            [
                feature.strip()
                for feature in json_decamelized["statistics_argument_names"]
            ]
            if json_decamelized.get("statistics_argument_names", None)
            else None
        )

        output_column_names = (
            [feature.strip() for feature in json_decamelized["output_column_names"]]
            if json_decamelized.get("output_column_names", None)
            else None
        )

        # Reconstructing statistics arguments.
        arg_list, _, _, _ = HopsworksUdf._parse_function_signature(function_source_code)

        transformation_features = (
            arg_list if not transformation_features else transformation_features
        )

        dropped_feature_names = (
            [
                transformation_features[arg_list.index(dropped_argument_name)]
                for dropped_argument_name in dropped_argument_names
            ]
            if dropped_argument_names
            else None
        )

        if statistics_features:
            transformation_features = [
                TransformationFeature(
                    transformation_features[arg_index],
                    arg_list[arg_index]
                    if arg_list[arg_index] in statistics_features
                    else None,
                )
                for arg_index in range(len(arg_list))
            ]
        else:
            transformation_features = [
                TransformationFeature(transformation_features[arg_index], None)
                for arg_index in range(len(arg_list))
            ]

        hopsworks_udf: HopsworksUdf = cls(
            func=function_source_code,
            return_types=output_types,
            name=function_name,
            transformation_features=transformation_features,
            dropped_argument_names=dropped_argument_names,
            dropped_feature_names=dropped_feature_names,
            feature_name_prefix=feature_name_prefix,
            transformation_function_argument_names=transformation_function_argument_names,
            execution_mode=UDFExecutionMode.from_string(
                json_decamelized["execution_mode"]
            ),
            output_column_names=output_column_names,
            generate_output_col_names=not output_column_names,  # Do not generate output column names if they are retrieved from the back
        )

        # Set transformation features if already set.
        return hopsworks_udf

    @property
    def return_types(self) -> List[str]:
        """Get the output types of the UDF"""
        # Update the number of outputs for one hot encoder to match the number of unique values for the feature
        if self.function_name == "one_hot_encoder" and self.transformation_statistics:
            self.update_return_type_one_hot()
        return self._return_types

    @property
    def function_name(self) -> str:
        """Get the function name of the UDF"""
        return self._function_name

    @property
    def statistics_required(self) -> bool:
        """Get if statistics for any feature is required by the UDF"""
        return bool(self.statistics_features)

    @property
    def transformation_statistics(
        self,
    ) -> Optional[TransformationStatistics]:
        """Feature statistics required for the defined UDF"""
        return self._statistics

    @property
    def output_column_names(self) -> List[str]:
        """Output columns names of the transformation function"""
        if self._feature_name_prefix:
            return [
                self._feature_name_prefix + output_col_name
                for output_col_name in self._output_column_names
            ]
        else:
            return self._output_column_names

    @property
    def transformation_features(self) -> List[str]:
        """
        List of feature names to be used in the User Defined Function.
        """
        if self._feature_name_prefix:
            return [
                self._feature_name_prefix + transformation_feature.feature_name
                for transformation_feature in self._transformation_features
            ]

        else:
            return [
                transformation_feature.feature_name
                for transformation_feature in self._transformation_features
            ]

    @property
    def unprefixed_transformation_features(self) -> List[str]:
        """
        List of feature name used in the transformation function without the feature name prefix.
        """
        return [
            transformation_feature.feature_name
            for transformation_feature in self._transformation_features
        ]

    @property
    def feature_name_prefix(self) -> Optional[str]:
        """The feature name prefix that needs to be added to the feature names"""
        return self._feature_name_prefix

    @property
    def statistics_features(self) -> List[str]:
        """
        List of feature names that require statistics
        """
        return [
            transformation_feature.feature_name
            for transformation_feature in self._transformation_features
            if transformation_feature.statistic_argument_name is not None
        ]

    @property
    def _statistics_argument_mapping(self) -> Dict[str, str]:
        """
        Dictionary that maps feature names to the statistics arguments names in the User defined function.
        """
        return {
            transformation_feature.feature_name: transformation_feature.statistic_argument_name
            for transformation_feature in self._transformation_features
        }

    @property
    def _statistics_argument_names(self) -> List[str]:
        """
        List of argument names required for statistics
        """
        return [
            transformation_feature.statistic_argument_name
            for transformation_feature in self._transformation_features
            if transformation_feature.statistic_argument_name is not None
        ]

    @property
    def dropped_features(self) -> List[str]:
        """
        List of features that will be dropped after the UDF is applied.
        """
        if self._feature_name_prefix and self._dropped_features:
            return [
                self._feature_name_prefix + dropped_feature
                for dropped_feature in self._dropped_features
            ]
        else:
            return self._dropped_features

    @property
    def execution_mode(self) -> UDFExecutionMode:
        return self._execution_mode

    @property
    def transformation_context(self) -> Dict[str, Any]:
        """
        Dictionary that contains the context variables required for the UDF.
        These context variables passed to the UDF during execution.
        """
        return self._transformation_context if self._transformation_context else {}

    @transformation_context.setter
    def transformation_context(self, context_variables: Dict[str, Any]) -> None:
        self._transformation_context = (
            self._validate_transformation_context(context_variables)
            if context_variables
            else {}
        )

    @dropped_features.setter
    def dropped_features(self, features: List[str]) -> None:
        self._dropped_features = HopsworksUdf._validate_and_convert_drop_features(
            features, self.transformation_features, self._feature_name_prefix
        )

    @transformation_statistics.setter
    def transformation_statistics(
        self, statistics: List[FeatureDescriptiveStatistics]
    ) -> None:
        self._statistics = TransformationStatistics(*self._statistics_argument_names)
        for stat in statistics:
            if stat.feature_name in self._statistics_argument_mapping.keys():
                self._statistics.set_statistics(
                    self._statistics_argument_mapping[stat.feature_name], stat.to_dict()
                )

    @output_column_names.setter
    def output_column_names(self, output_col_names: Union[str, List[str]]) -> None:
        if not isinstance(output_col_names, List):
            output_col_names = [output_col_names]
        self._validate_output_col_name(output_col_names)
        self._output_column_names = output_col_names

    def __repr__(self):
        return f'{self.function_name}({", ".join(self.transformation_features)})'



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import List, Optional

from hopsworks_common import client, decorators, environment, usage
from hopsworks_common.engine import environment_engine


class EnvironmentApi:
    def __init__(self):
        self._environment_engine = environment_engine.EnvironmentEngine()

    @usage.method_logger
    def create_environment(
        self,
        name: str,
        description: Optional[str] = None,
        base_environment_name: Optional[str] = "python-feature-pipeline",
        await_creation: Optional[bool] = True,
    ) -> environment.Environment:
        """Create Python environment for the project

        ```python

        import hopsworks

        project = hopsworks.login()

        env_api = project.get_environment_api()

        new_env = env_api.create_environment("my_custom_environment", base_environment_name="python-feature-pipeline")


        ```
        # Arguments
            name: name of the environment
            base_environment_name: the name of the environment to clone from
            await_creation: bool. If True the method returns only when the creation is finished. Default True
        # Returns
            `Environment`: The Environment object
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()

        path_params = [
            "project",
            _client._project_id,
            "python",
            "environments",
            name,
        ]
        headers = {"content-type": "application/json"}
        data = {
            "name": name,
            "baseImage": {"name": base_environment_name, "description": description},
        }
        env = environment.Environment.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, data=json.dumps(data)
            )
        )

        if await_creation:
            self._environment_engine.await_environment_command(name)

        return env

    @usage.method_logger
    def get_environments(self) -> List[environment.Environment]:
        """Get all available environments in the project.

        ```python

        import hopsworks

        project = hopsworks.login()

        env_api = project.get_environment_api()

        envs = env_api.get_environments()

        ```
        # Returns
            `List[Environment]`: List of Environment objects
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()

        path_params = ["project", _client._project_id, "python", "environments"]
        query_params = {"expand": ["libraries", "commands"]}
        headers = {"content-type": "application/json"}
        return environment.Environment.from_response_json(
            _client._send_request(
                "GET", path_params, query_params=query_params, headers=headers
            )
        )

    @usage.method_logger
    @decorators.catch_not_found(
        "hopsworks_common.environment.Environment", fallback_return=None
    )
    def get_environment(self, name: str) -> Optional[environment.Environment]:
        """Get handle for a Python environment in the project

        ```python

        import hopsworks

        project = hopsworks.login()

        env_api = project.get_environment_api()

        env = env_api.get_environment("my_custom_environment")

        ```
        # Arguments
            name: name of the environment
        # Returns
            `Environment`: The Environment object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()

        path_params = ["project", _client._project_id, "python", "environments", name]
        query_params = {"expand": ["libraries", "commands"]}
        headers = {"content-type": "application/json"}
        return environment.Environment.from_response_json(
            _client._send_request(
                "GET", path_params, query_params=query_params, headers=headers
            )
        )

    def _delete(self, name):
        """Delete the Python environment.
        :param name: name of environment to delete
        :type environment: Environment
        """
        _client = client.get_instance()

        path_params = [
            "project",
            _client._project_id,
            "python",
            "environments",
            name,
        ]
        headers = {"content-type": "application/json"}
        (_client._send_request("DELETE", path_params, headers=headers),)



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import Any, Dict, Optional, Union

from hsfs import feature, util


class Filter:
    GE = "GREATER_THAN_OR_EQUAL"
    GT = "GREATER_THAN"
    NE = "NOT_EQUALS"
    EQ = "EQUALS"
    LE = "LESS_THAN_OR_EQUAL"
    LT = "LESS_THAN"
    IN = "IN"
    LK = "LIKE"

    def __init__(
        self, feature: "feature.Feature", condition: str, value: Any, **kwargs
    ) -> None:
        self._feature = feature
        self._condition = condition
        self._value = value

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> Optional["Filter"]:
        if json_dict is None:
            return None

        return cls(
            feature=feature.Feature.from_response_json(json_dict["feature"]),
            condition=json_dict["condition"],
            value=json_dict["value"],
        )

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "feature": self._feature,
            "condition": self._condition,
            "value": str(self._value) if self._value is not None else None,
        }

    def __and__(self, other: Union["Logic", "Filter"]) -> "Logic":
        if isinstance(other, Filter):
            return Logic.And(left_f=self, right_f=other)
        elif isinstance(other, Logic):
            return Logic.And(left_f=self, right_l=other)
        else:
            raise TypeError(
                "Operator `&` expected type `Filter` or `Logic`, got `{}`".format(
                    type(other)
                )
            )

    def __or__(self, other: Union["Filter", "Logic"]) -> "Logic":
        if isinstance(other, Filter):
            return Logic.Or(left_f=self, right_f=other)
        elif isinstance(other, Logic):
            return Logic.Or(left_f=self, right_l=other)
        else:
            raise TypeError(
                "Operator `|` expected type `Filter` or `Logic`, got `{}`".format(
                    type(other)
                )
            )

    def __repr__(self) -> str:
        return f"Filter({self._feature!r}, {self._condition!r}, {self._value!r})"

    def __str__(self) -> str:
        return self.json()

    @property
    def feature(self) -> "feature.Feature":
        return self._feature

    @property
    def condition(self) -> str:
        return self._condition

    @property
    def value(self) -> Any:
        return self._value


class Logic:
    AND = "AND"
    OR = "OR"
    SINGLE = "SINGLE"

    def __init__(
        self,
        type: Optional[str],
        left_f: Optional["Filter"] = None,
        right_f: Optional["Filter"] = None,
        left_l: Optional["Logic"] = None,
        right_l: Optional["Logic"] = None,
        **kwargs,
    ) -> None:
        self._type = type
        self._left_f = left_f
        self._right_f = right_f
        self._left_l = left_l
        self._right_l = right_l

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self._type,
            "leftFilter": self._left_f,
            "rightFilter": self._right_f,
            "leftLogic": self._left_l,
            "rightLogic": self._right_l,
        }

    @classmethod
    def from_response_json(
        cls, json_dict: Optional[Dict[str, Any]]
    ) -> Optional["Logic"]:
        if json_dict is None:
            return None

        return cls(
            type=json_dict["type"] if "type" in json_dict else None,
            left_f=Filter.from_response_json(json_dict.get("left_filter", None)),
            right_f=Filter.from_response_json(json_dict.get("right_filter", None)),
            left_l=Logic.from_response_json(json_dict.get("left_logic", None)),
            right_l=Logic.from_response_json(json_dict.get("right_logic", None)),
        )

    @classmethod
    def And(
        cls,
        left_f: Optional["Filter"] = None,
        right_f: Optional["Filter"] = None,
        left_l: Optional["Logic"] = None,
        right_l: Optional["Logic"] = None,
    ) -> "Logic":
        return cls(cls.AND, left_f, right_f, left_l, right_l)

    @classmethod
    def Or(
        cls,
        left_f: Optional["Filter"] = None,
        right_f: Optional["Filter"] = None,
        left_l: Optional["Logic"] = None,
        right_l: Optional["Logic"] = None,
    ) -> Logic:
        return cls(cls.OR, left_f, right_f, left_l, right_l)

    @classmethod
    def Single(cls, left_f: "Filter") -> Logic:
        return cls(cls.SINGLE, left_f)

    def __and__(self, other: Union["Filter", "Logic"]) -> "Logic":
        if isinstance(other, Filter):
            return Logic.And(left_l=self, right_f=other)
        elif isinstance(other, Logic):
            return Logic.And(left_l=self, right_l=other)
        else:
            raise TypeError(
                "Operator `&` expected type `Filter` or `Logic`, got `{}`".format(
                    type(other)
                )
            )

    def __or__(self, other: "Logic") -> "Logic":
        if isinstance(other, Filter):
            return Logic.Or(left_l=self, right_f=other)
        elif isinstance(other, Logic):
            return Logic.Or(left_l=self, right_l=other)
        else:
            raise TypeError(
                "Operator `|` expected type `Filter` or `Logic`, got `{}`".format(
                    type(other)
                )
            )

    def __repr__(self) -> str:
        return f"Logic({self._type!r}, {self._left_f!r}, {self._right_f!r}, {self._left_l!r}, {self._right_l!r})"

    def __str__(self) -> str:
        return self.json()

    @property
    def type(self) -> Optional[str]:
        return self._type

    def get_left_filter_or_logic(self) -> Union["Filter", "Logic"]:
        return self._left_f or self._left_l

    def get_right_filter_or_logic(self) -> Optional[Union["Filter", "Logic"]]:
        return self._right_f or self._right_l



#
#   Copyright 2023 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from dataclasses import dataclass
from typing import List, Optional

import humps
from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException
from hsfs import util
from hsfs.core.vector_db_client import VectorDbClient


class SimilarityFunctionType:
    """
    Enumeration class representing different types of similarity functions.

    # Attributes
        L2 (str): Represents L2 norm similarity function.
        COSINE (str): Represents cosine similarity function.
        DOT_PRODUCT (str): Represents dot product similarity function.
    """

    L2 = "l2_norm"
    COSINE = "cosine"
    DOT_PRODUCT = "dot_product"

    def __init__(self) -> None:
        # Fix for the doc
        raise NotImplementedError("This class should not be instantiated.")


@dataclass
class HsmlModel:
    """
    Data class storing the metadata of a hsml model
    """

    model_registry_id: int
    model_name: str
    model_version: int

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)

        return cls(
            model_registry_id=json_decamelized.get("model_registry_id"),
            model_name=json_decamelized.get("model_name"),
            model_version=json_decamelized.get("model_version"),
        )

    @classmethod
    def from_model(cls, model):
        return cls(
            model_registry_id=model.model_registry_id,
            model_name=model.name,
            model_version=model.version,
        )

    # should get from backend because of authorisation check (unshared project etc)
    def get_model(self):
        try:
            from hsml.model import Model
        except ModuleNotFoundError as err:
            raise FeatureStoreException(
                "Model is attached to embedding feature but hsml library is not installed."
                "Install hsml library before getting the feature group."
            ) from err

        path_params = [
            "project",
            client.get_instance()._project_id,
            "modelregistries",
            self.model_registry_id,
            "models",
            self.model_name + "_" + str(self.model_version),
        ]
        query_params = {"expand": "trainingdatasets"}

        model_json = client.get_instance()._send_request(
            "GET", path_params, query_params
        )
        return Model.from_response_json(model_json)

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "modelRegistryId": self.model_registry_id,
            "modelName": self.model_name,
            "modelVersion": self.model_version,
        }


class EmbeddingFeature:
    """
    Represents an embedding feature.

    # Arguments
        name: The name of the embedding feature.
        dimension: The dimensionality of the embedding feature.
        similarity_function_type: The type of similarity function used for the embedding feature.
          Available functions are `L2`, `COSINE`, and `DOT_PRODUCT`.
          (default is `SimilarityFunctionType.L2`).
        model: `hsml.model.Model` A Model in hsml.
        feature_group: The feature group object that contains the embedding feature.
        embedding_index: `EmbeddingIndex` The index for managing embedding features.
    """

    def __init__(
        self,
        name: str = None,
        dimension: int = None,
        similarity_function_type: SimilarityFunctionType = SimilarityFunctionType.L2,
        model=None,
        feature_group=None,
        embedding_index=None,
    ):
        self._name = name
        self._dimension = dimension
        self._similarity_function_type = similarity_function_type
        self._model = model
        self._feature_group = feature_group
        self._embedding_index = embedding_index

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        hsml_model_json = json_decamelized.get("model")
        hsml_model = (
            HsmlModel.from_response_json(hsml_model_json).get_model()
            if hsml_model_json
            else None
        )

        return cls(
            name=json_decamelized.get("name"),
            dimension=json_decamelized.get("dimension"),
            similarity_function_type=json_decamelized.get("similarity_function_type"),
            model=hsml_model,
        )

    @property
    def name(self):
        """str: The name of the embedding feature."""
        return self._name

    @property
    def dimenstion(self):
        """This one is excluded from the docs as the name is misspelled but kept to avoid breaking the API
        int: The dimensionality of the embedding feature."""
        return self._dimension

    @property
    def dimension(self):
        """int: The dimensionality of the embedding feature."""
        return self._dimension

    @property
    def similarity_function_type(self):
        """SimilarityFunctionType: The type of similarity function used for the embedding feature."""
        return self._similarity_function_type

    @property
    def model(self):
        """hsml.model.Model: The Model in hsml."""
        return self._model

    @property
    def feature_group(self):
        """FeatureGroup: The feature group object that contains the embedding feature."""
        return self._feature_group

    @feature_group.setter
    def feature_group(self, feature_group):
        """Set the feature group object that contains the embedding feature.

        Args:
            feature_group (FeatureGroup): The feature group object.
        """
        self._feature_group = feature_group

    @property
    def embedding_index(self):
        """EmbeddingIndex: The index for managing embedding features."""
        return self._embedding_index

    @embedding_index.setter
    def embedding_index(self, embedding_index):
        """Set the index for managing embedding features.

        Args:
            embedding_index (EmbeddingIndex): The embedding index object.
        """
        self._embedding_index = embedding_index

    def json(self):
        """Serialize the EmbeddingFeature object to a JSON string."""
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """
        Convert the EmbeddingFeature object to a dictionary.

        Returns:
            `dict`: A dictionary representation of the EmbeddingFeature object.
        """
        d = {
            "name": self._name,
            "dimension": self._dimension,
            "similarityFunctionType": self._similarity_function_type,
        }
        if self._model:
            d["model"] = HsmlModel.from_model(self._model)
        return d

    def __repr__(self):
        return self.json()


class EmbeddingIndex:
    """
    Represents an index for managing embedding features.

    # Arguments
        index_name: The name of the embedding index. The name of the project index is used if not provided.
        features: A list of `EmbeddingFeature` objects for the features that
            contain embeddings that should be indexed for similarity search.
        col_prefix: The prefix to be added to column names when using project index.
            It is managed by Hopsworks and should not be provided.

    !!! Example
        ```
        embedding_index = EmbeddingIndex()
        embedding_index.add_embedding(name="user_vector", dimension=256)
        embeddings = embedding_index.get_embeddings()
        ```
    """

    def __init__(
        self,
        index_name: Optional[str] = None,
        features: Optional[List[EmbeddingFeature]] = None,
        col_prefix: Optional[str] = None,
    ):
        self._index_name = index_name
        if features is None:
            self._features = {}
        else:
            self._features = dict([(feat.name, feat) for feat in features])
        self._feature_group = None
        self._col_prefix = col_prefix
        self._vector_db_client = None

    def add_embedding(
        self,
        name: str,
        dimension: int,
        similarity_function_type: Optional[
            SimilarityFunctionType
        ] = SimilarityFunctionType.L2,
        model=None,
    ):
        """
        Adds a new embedding feature to the index.

        Example:
        ```
        embedding_index = EmbeddingIndex()
        embedding_index.add_embedding(name="user_vector", dimension=256)

        # Attach a hsml model to the embedding feature
        embedding_index = EmbeddingIndex()
        embedding_index.add_embedding(name="user_vector", dimension=256, model=hsml_model)
        ```

        # Arguments
            name: The name of the embedding feature.
            dimension: The dimensionality of the embedding feature.
            similarity_function_type: The type of similarity function to be used.
            model (hsml.model.Model, optional): The hsml model used to generate the embedding.
                Defaults to None.
        """
        self._features[name] = EmbeddingFeature(
            name, dimension, similarity_function_type, model=model
        )

    def get_embedding(self, name):
        """
        Returns the `hsfs.embedding.EmbeddingFeature` object associated with the feature name.

        # Arguments
            name (str): The name of the embedding feature.

        # Returns
            `hsfs.embedding.EmbeddingFeature` object
        """
        feat = self._features.get(name)
        if feat:
            feat.feature_group = self._feature_group
            feat.embedding_index = self
        return feat

    def get_embeddings(self):
        """
        Returns the list of `hsfs.embedding.EmbeddingFeature` objects associated with the index.

        # Returns
            A list of `hsfs.embedding.EmbeddingFeature` objects
        """
        for feat in self._features.values():
            feat.feature_group = self._feature_group
            feat.embedding_index = self
        return self._features.values()

    def count(self, options: map = None):
        """
        Count the number of records in the feature group.

        # Arguments
            options: The options used for the request to the vector database.
                The keys are attribute values of the `hsfs.core.opensearch.OpensearchRequestOption` class.

        # Returns
            `int`: The number of records in the feature group.

        # Raises
            `ValueError`: If the feature group is not initialized.
            `hopsworks.client.exceptions.FeatureStoreException`: If an error occurs during the count operation.
        """
        if self._vector_db_client is None:
            self._vector_db_client = VectorDbClient(self._feature_group.select_all())
        return self._vector_db_client.count(self.feature_group, options=options)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(
            index_name=json_decamelized.get("index_name"),
            features=[
                EmbeddingFeature.from_response_json(f)
                for f in json_decamelized.get("features")
            ],
            col_prefix=json_decamelized.get("col_prefix"),
        )

    @property
    def feature_group(self):
        """FeatureGroup: The feature group object that contains the embedding feature."""
        return self._feature_group

    @feature_group.setter
    def feature_group(self, feature_group):
        """Setter for the feature group object."""
        self._feature_group = feature_group

    @property
    def index_name(self):
        """str: The name of the embedding index."""
        return self._index_name

    @property
    def col_prefix(self):
        """str: The prefix to be added to column names."""
        return self._col_prefix

    def json(self):
        """Serialize the EmbeddingIndex object to a JSON string."""
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        """
        Convert the EmbeddingIndex object to a dictionary.

        Returns:
            `dict`: A dictionary representation of the EmbeddingIndex object.
        """
        return {
            "indexName": self._index_name,
            "features": [feature.to_dict() for feature in self._features.values()],
            "colPrefix": self._col_prefix,
        }

    def __repr__(self):
        return self.json()



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from hopsworks_common import client
from hsfs.core import data_source as ds
from hsfs.core import data_source_data as dsd


class DataSourceApi:

    def get_databases(self, feature_store_id: int, name: str) -> list[str]:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            name,
            "data_source",
            "databases",
        ]

        return _client._send_request("GET", path_params)

    def get_tables(self, feature_store_id: int, name: str, database: str) -> list[ds.DataSource]:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            name,
            "data_source",
            "tables",
        ]

        query_params = {"database": database}

        return ds.DataSource.from_response_json(
            _client._send_request("GET", path_params, query_params)
        )

    def get_data(self, feature_store_id: int, name: str, data_source: ds.DataSource) -> dsd.DataSourceData:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            name,
            "data_source",
            "data",
        ]

        query_params = data_source.to_dict()

        return dsd.DataSourceData.from_response_json(
            _client._send_request("GET", path_params, query_params)
        )


    def get_metadata(self, feature_store_id: int, name: str, data_source: ds.DataSource) -> dict:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "storageconnectors",
            name,
            "data_source",
            "metadata",
        ]

        query_params = data_source.to_dict()

        return _client._send_request("GET", path_params, query_params)



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from typing import Any, Dict, Optional

import humps
from hopsworks_common import util


@dataclass
class User:
    email: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    href: Optional[str] = None
    username: Optional[str] = None
    status: Optional[str] = None
    secret: Optional[str] = None
    chosen_password: Optional[str] = None
    repeated_password: Optional[str] = None
    tos: Optional[bool] = None
    two_factor: Optional[bool] = None
    user_account_type: Optional[str] = None
    tours_state: Optional[bool] = None
    max_num_projects: Optional[int] = None
    test_user: Optional[bool] = None
    num_active_projects: Optional[int] = None
    num_remaining_projects: Optional[int] = None

    @classmethod
    def from_response_json(cls, json_dict: Optional[Dict[str, Any]]) -> Optional[User]:
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            if "firstname" in json_decamelized.keys():
                json_decamelized["first_name"] = json_decamelized.pop("firstname")
                json_decamelized["last_name"] = json_decamelized.pop("lastname")
            # Remove keys that are not part of the dataclass
            for key in set(json_decamelized.keys()) - set(
                User.__dataclass_fields__.keys()
            ):
                json_decamelized.pop(key)
            return cls(**json_decamelized)
        else:
            return None

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc_predict_v2.proto
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15grpc_predict_v2.proto\x12\tinference"\x13\n\x11ServerLiveRequest""\n\x12ServerLiveResponse\x12\x0c\n\x04live\x18\x01 \x01(\x08"\x14\n\x12ServerReadyRequest"$\n\x13ServerReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"2\n\x11ModelReadyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t"#\n\x12ModelReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"\x17\n\x15ServerMetadataRequest"K\n\x16ServerMetadataResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x12\n\nextensions\x18\x03 \x03(\t"5\n\x14ModelMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t"\x8d\x02\n\x15ModelMetadataResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08versions\x18\x02 \x03(\t\x12\x10\n\x08platform\x18\x03 \x01(\t\x12?\n\x06inputs\x18\x04 \x03(\x0b\x32/.inference.ModelMetadataResponse.TensorMetadata\x12@\n\x07outputs\x18\x05 \x03(\x0b\x32/.inference.ModelMetadataResponse.TensorMetadata\x1a?\n\x0eTensorMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03"\xee\x06\n\x11ModelInferRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12@\n\nparameters\x18\x04 \x03(\x0b\x32,.inference.ModelInferRequest.ParametersEntry\x12=\n\x06inputs\x18\x05 \x03(\x0b\x32-.inference.ModelInferRequest.InferInputTensor\x12H\n\x07outputs\x18\x06 \x03(\x0b\x32\x37.inference.ModelInferRequest.InferRequestedOutputTensor\x12\x1a\n\x12raw_input_contents\x18\x07 \x03(\x0c\x1a\x94\x02\n\x10InferInputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\x12Q\n\nparameters\x18\x04 \x03(\x0b\x32=.inference.ModelInferRequest.InferInputTensor.ParametersEntry\x12\x30\n\x08\x63ontents\x18\x05 \x01(\x0b\x32\x1e.inference.InferTensorContents\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1a\xd5\x01\n\x1aInferRequestedOutputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12[\n\nparameters\x18\x02 \x03(\x0b\x32G.inference.ModelInferRequest.InferRequestedOutputTensor.ParametersEntry\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01"\xd5\x04\n\x12ModelInferResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12\x41\n\nparameters\x18\x04 \x03(\x0b\x32-.inference.ModelInferResponse.ParametersEntry\x12@\n\x07outputs\x18\x05 \x03(\x0b\x32/.inference.ModelInferResponse.InferOutputTensor\x12\x1b\n\x13raw_output_contents\x18\x06 \x03(\x0c\x1a\x97\x02\n\x11InferOutputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\x12S\n\nparameters\x18\x04 \x03(\x0b\x32?.inference.ModelInferResponse.InferOutputTensor.ParametersEntry\x12\x30\n\x08\x63ontents\x18\x05 \x01(\x0b\x32\x1e.inference.InferTensorContents\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01"i\n\x0eInferParameter\x12\x14\n\nbool_param\x18\x01 \x01(\x08H\x00\x12\x15\n\x0bint64_param\x18\x02 \x01(\x03H\x00\x12\x16\n\x0cstring_param\x18\x03 \x01(\tH\x00\x42\x12\n\x10parameter_choice"\xd0\x01\n\x13InferTensorContents\x12\x15\n\rbool_contents\x18\x01 \x03(\x08\x12\x14\n\x0cint_contents\x18\x02 \x03(\x05\x12\x16\n\x0eint64_contents\x18\x03 \x03(\x03\x12\x15\n\ruint_contents\x18\x04 \x03(\r\x12\x17\n\x0fuint64_contents\x18\x05 \x03(\x04\x12\x15\n\rfp32_contents\x18\x06 \x03(\x02\x12\x15\n\rfp64_contents\x18\x07 \x03(\x01\x12\x16\n\x0e\x62ytes_contents\x18\x08 \x03(\x0c"0\n\x1aRepositoryModelLoadRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t"C\n\x1bRepositoryModelLoadResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x10\n\x08isLoaded\x18\x02 \x01(\x08"2\n\x1cRepositoryModelUnloadRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t"G\n\x1dRepositoryModelUnloadResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x12\n\nisUnloaded\x18\x02 \x01(\x08\x32\xd2\x05\n\x14GRPCInferenceService\x12K\n\nServerLive\x12\x1c.inference.ServerLiveRequest\x1a\x1d.inference.ServerLiveResponse"\x00\x12N\n\x0bServerReady\x12\x1d.inference.ServerReadyRequest\x1a\x1e.inference.ServerReadyResponse"\x00\x12K\n\nModelReady\x12\x1c.inference.ModelReadyRequest\x1a\x1d.inference.ModelReadyResponse"\x00\x12W\n\x0eServerMetadata\x12 .inference.ServerMetadataRequest\x1a!.inference.ServerMetadataResponse"\x00\x12T\n\rModelMetadata\x12\x1f.inference.ModelMetadataRequest\x1a .inference.ModelMetadataResponse"\x00\x12K\n\nModelInfer\x12\x1c.inference.ModelInferRequest\x1a\x1d.inference.ModelInferResponse"\x00\x12\x66\n\x13RepositoryModelLoad\x12%.inference.RepositoryModelLoadRequest\x1a&.inference.RepositoryModelLoadResponse"\x00\x12l\n\x15RepositoryModelUnload\x12\'.inference.RepositoryModelUnloadRequest\x1a(.inference.RepositoryModelUnloadResponse"\x00\x62\x06proto3'
)


_SERVERLIVEREQUEST = DESCRIPTOR.message_types_by_name["ServerLiveRequest"]
_SERVERLIVERESPONSE = DESCRIPTOR.message_types_by_name["ServerLiveResponse"]
_SERVERREADYREQUEST = DESCRIPTOR.message_types_by_name["ServerReadyRequest"]
_SERVERREADYRESPONSE = DESCRIPTOR.message_types_by_name["ServerReadyResponse"]
_MODELREADYREQUEST = DESCRIPTOR.message_types_by_name["ModelReadyRequest"]
_MODELREADYRESPONSE = DESCRIPTOR.message_types_by_name["ModelReadyResponse"]
_SERVERMETADATAREQUEST = DESCRIPTOR.message_types_by_name["ServerMetadataRequest"]
_SERVERMETADATARESPONSE = DESCRIPTOR.message_types_by_name["ServerMetadataResponse"]
_MODELMETADATAREQUEST = DESCRIPTOR.message_types_by_name["ModelMetadataRequest"]
_MODELMETADATARESPONSE = DESCRIPTOR.message_types_by_name["ModelMetadataResponse"]
_MODELMETADATARESPONSE_TENSORMETADATA = _MODELMETADATARESPONSE.nested_types_by_name[
    "TensorMetadata"
]
_MODELINFERREQUEST = DESCRIPTOR.message_types_by_name["ModelInferRequest"]
_MODELINFERREQUEST_INFERINPUTTENSOR = _MODELINFERREQUEST.nested_types_by_name[
    "InferInputTensor"
]
_MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY = (
    _MODELINFERREQUEST_INFERINPUTTENSOR.nested_types_by_name["ParametersEntry"]
)
_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR = _MODELINFERREQUEST.nested_types_by_name[
    "InferRequestedOutputTensor"
]
_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY = (
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR.nested_types_by_name[
        "ParametersEntry"
    ]
)
_MODELINFERREQUEST_PARAMETERSENTRY = _MODELINFERREQUEST.nested_types_by_name[
    "ParametersEntry"
]
_MODELINFERRESPONSE = DESCRIPTOR.message_types_by_name["ModelInferResponse"]
_MODELINFERRESPONSE_INFEROUTPUTTENSOR = _MODELINFERRESPONSE.nested_types_by_name[
    "InferOutputTensor"
]
_MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY = (
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR.nested_types_by_name["ParametersEntry"]
)
_MODELINFERRESPONSE_PARAMETERSENTRY = _MODELINFERRESPONSE.nested_types_by_name[
    "ParametersEntry"
]
_INFERPARAMETER = DESCRIPTOR.message_types_by_name["InferParameter"]
_INFERTENSORCONTENTS = DESCRIPTOR.message_types_by_name["InferTensorContents"]
_REPOSITORYMODELLOADREQUEST = DESCRIPTOR.message_types_by_name[
    "RepositoryModelLoadRequest"
]
_REPOSITORYMODELLOADRESPONSE = DESCRIPTOR.message_types_by_name[
    "RepositoryModelLoadResponse"
]
_REPOSITORYMODELUNLOADREQUEST = DESCRIPTOR.message_types_by_name[
    "RepositoryModelUnloadRequest"
]
_REPOSITORYMODELUNLOADRESPONSE = DESCRIPTOR.message_types_by_name[
    "RepositoryModelUnloadResponse"
]
ServerLiveRequest = _reflection.GeneratedProtocolMessageType(
    "ServerLiveRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERLIVEREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ServerLiveRequest)
    },
)
_sym_db.RegisterMessage(ServerLiveRequest)

ServerLiveResponse = _reflection.GeneratedProtocolMessageType(
    "ServerLiveResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERLIVERESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ServerLiveResponse)
    },
)
_sym_db.RegisterMessage(ServerLiveResponse)

ServerReadyRequest = _reflection.GeneratedProtocolMessageType(
    "ServerReadyRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERREADYREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ServerReadyRequest)
    },
)
_sym_db.RegisterMessage(ServerReadyRequest)

ServerReadyResponse = _reflection.GeneratedProtocolMessageType(
    "ServerReadyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERREADYRESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ServerReadyResponse)
    },
)
_sym_db.RegisterMessage(ServerReadyResponse)

ModelReadyRequest = _reflection.GeneratedProtocolMessageType(
    "ModelReadyRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELREADYREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ModelReadyRequest)
    },
)
_sym_db.RegisterMessage(ModelReadyRequest)

ModelReadyResponse = _reflection.GeneratedProtocolMessageType(
    "ModelReadyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELREADYRESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ModelReadyResponse)
    },
)
_sym_db.RegisterMessage(ModelReadyResponse)

ServerMetadataRequest = _reflection.GeneratedProtocolMessageType(
    "ServerMetadataRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERMETADATAREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ServerMetadataRequest)
    },
)
_sym_db.RegisterMessage(ServerMetadataRequest)

ServerMetadataResponse = _reflection.GeneratedProtocolMessageType(
    "ServerMetadataResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERMETADATARESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ServerMetadataResponse)
    },
)
_sym_db.RegisterMessage(ServerMetadataResponse)

ModelMetadataRequest = _reflection.GeneratedProtocolMessageType(
    "ModelMetadataRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELMETADATAREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ModelMetadataRequest)
    },
)
_sym_db.RegisterMessage(ModelMetadataRequest)

ModelMetadataResponse = _reflection.GeneratedProtocolMessageType(
    "ModelMetadataResponse",
    (_message.Message,),
    {
        "TensorMetadata": _reflection.GeneratedProtocolMessageType(
            "TensorMetadata",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELMETADATARESPONSE_TENSORMETADATA,
                "__module__": "grpc_predict_v2_pb2",
                # @@protoc_insertion_point(class_scope:inference.ModelMetadataResponse.TensorMetadata)
            },
        ),
        "DESCRIPTOR": _MODELMETADATARESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ModelMetadataResponse)
    },
)
_sym_db.RegisterMessage(ModelMetadataResponse)
_sym_db.RegisterMessage(ModelMetadataResponse.TensorMetadata)

ModelInferRequest = _reflection.GeneratedProtocolMessageType(
    "ModelInferRequest",
    (_message.Message,),
    {
        "InferInputTensor": _reflection.GeneratedProtocolMessageType(
            "InferInputTensor",
            (_message.Message,),
            {
                "ParametersEntry": _reflection.GeneratedProtocolMessageType(
                    "ParametersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY,
                        "__module__": "grpc_predict_v2_pb2",
                        # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferInputTensor.ParametersEntry)
                    },
                ),
                "DESCRIPTOR": _MODELINFERREQUEST_INFERINPUTTENSOR,
                "__module__": "grpc_predict_v2_pb2",
                # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferInputTensor)
            },
        ),
        "InferRequestedOutputTensor": _reflection.GeneratedProtocolMessageType(
            "InferRequestedOutputTensor",
            (_message.Message,),
            {
                "ParametersEntry": _reflection.GeneratedProtocolMessageType(
                    "ParametersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY,
                        "__module__": "grpc_predict_v2_pb2",
                        # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferRequestedOutputTensor.ParametersEntry)
                    },
                ),
                "DESCRIPTOR": _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR,
                "__module__": "grpc_predict_v2_pb2",
                # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.InferRequestedOutputTensor)
            },
        ),
        "ParametersEntry": _reflection.GeneratedProtocolMessageType(
            "ParametersEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELINFERREQUEST_PARAMETERSENTRY,
                "__module__": "grpc_predict_v2_pb2",
                # @@protoc_insertion_point(class_scope:inference.ModelInferRequest.ParametersEntry)
            },
        ),
        "DESCRIPTOR": _MODELINFERREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ModelInferRequest)
    },
)
_sym_db.RegisterMessage(ModelInferRequest)
_sym_db.RegisterMessage(ModelInferRequest.InferInputTensor)
_sym_db.RegisterMessage(ModelInferRequest.InferInputTensor.ParametersEntry)
_sym_db.RegisterMessage(ModelInferRequest.InferRequestedOutputTensor)
_sym_db.RegisterMessage(ModelInferRequest.InferRequestedOutputTensor.ParametersEntry)
_sym_db.RegisterMessage(ModelInferRequest.ParametersEntry)

ModelInferResponse = _reflection.GeneratedProtocolMessageType(
    "ModelInferResponse",
    (_message.Message,),
    {
        "InferOutputTensor": _reflection.GeneratedProtocolMessageType(
            "InferOutputTensor",
            (_message.Message,),
            {
                "ParametersEntry": _reflection.GeneratedProtocolMessageType(
                    "ParametersEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY,
                        "__module__": "grpc_predict_v2_pb2",
                        # @@protoc_insertion_point(class_scope:inference.ModelInferResponse.InferOutputTensor.ParametersEntry)
                    },
                ),
                "DESCRIPTOR": _MODELINFERRESPONSE_INFEROUTPUTTENSOR,
                "__module__": "grpc_predict_v2_pb2",
                # @@protoc_insertion_point(class_scope:inference.ModelInferResponse.InferOutputTensor)
            },
        ),
        "ParametersEntry": _reflection.GeneratedProtocolMessageType(
            "ParametersEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELINFERRESPONSE_PARAMETERSENTRY,
                "__module__": "grpc_predict_v2_pb2",
                # @@protoc_insertion_point(class_scope:inference.ModelInferResponse.ParametersEntry)
            },
        ),
        "DESCRIPTOR": _MODELINFERRESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.ModelInferResponse)
    },
)
_sym_db.RegisterMessage(ModelInferResponse)
_sym_db.RegisterMessage(ModelInferResponse.InferOutputTensor)
_sym_db.RegisterMessage(ModelInferResponse.InferOutputTensor.ParametersEntry)
_sym_db.RegisterMessage(ModelInferResponse.ParametersEntry)

InferParameter = _reflection.GeneratedProtocolMessageType(
    "InferParameter",
    (_message.Message,),
    {
        "DESCRIPTOR": _INFERPARAMETER,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.InferParameter)
    },
)
_sym_db.RegisterMessage(InferParameter)

InferTensorContents = _reflection.GeneratedProtocolMessageType(
    "InferTensorContents",
    (_message.Message,),
    {
        "DESCRIPTOR": _INFERTENSORCONTENTS,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.InferTensorContents)
    },
)
_sym_db.RegisterMessage(InferTensorContents)

RepositoryModelLoadRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelLoadRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELLOADREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelLoadRequest)
    },
)
_sym_db.RegisterMessage(RepositoryModelLoadRequest)

RepositoryModelLoadResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelLoadResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELLOADRESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelLoadResponse)
    },
)
_sym_db.RegisterMessage(RepositoryModelLoadResponse)

RepositoryModelUnloadRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelUnloadRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELUNLOADREQUEST,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelUnloadRequest)
    },
)
_sym_db.RegisterMessage(RepositoryModelUnloadRequest)

RepositoryModelUnloadResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelUnloadResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELUNLOADRESPONSE,
        "__module__": "grpc_predict_v2_pb2",
        # @@protoc_insertion_point(class_scope:inference.RepositoryModelUnloadResponse)
    },
)
_sym_db.RegisterMessage(RepositoryModelUnloadResponse)

_GRPCINFERENCESERVICE = DESCRIPTOR.services_by_name["GRPCInferenceService"]
if _descriptor._USE_C_DESCRIPTORS == False:  # noqa: E712
    DESCRIPTOR._options = None
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._options = None
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._options = None
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERREQUEST_PARAMETERSENTRY._options = None
    _MODELINFERREQUEST_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._options = None
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._serialized_options = b"8\001"
    _MODELINFERRESPONSE_PARAMETERSENTRY._options = None
    _MODELINFERRESPONSE_PARAMETERSENTRY._serialized_options = b"8\001"
    _SERVERLIVEREQUEST._serialized_start = 36
    _SERVERLIVEREQUEST._serialized_end = 55
    _SERVERLIVERESPONSE._serialized_start = 57
    _SERVERLIVERESPONSE._serialized_end = 91
    _SERVERREADYREQUEST._serialized_start = 93
    _SERVERREADYREQUEST._serialized_end = 113
    _SERVERREADYRESPONSE._serialized_start = 115
    _SERVERREADYRESPONSE._serialized_end = 151
    _MODELREADYREQUEST._serialized_start = 153
    _MODELREADYREQUEST._serialized_end = 203
    _MODELREADYRESPONSE._serialized_start = 205
    _MODELREADYRESPONSE._serialized_end = 240
    _SERVERMETADATAREQUEST._serialized_start = 242
    _SERVERMETADATAREQUEST._serialized_end = 265
    _SERVERMETADATARESPONSE._serialized_start = 267
    _SERVERMETADATARESPONSE._serialized_end = 342
    _MODELMETADATAREQUEST._serialized_start = 344
    _MODELMETADATAREQUEST._serialized_end = 397
    _MODELMETADATARESPONSE._serialized_start = 400
    _MODELMETADATARESPONSE._serialized_end = 669
    _MODELMETADATARESPONSE_TENSORMETADATA._serialized_start = 606
    _MODELMETADATARESPONSE_TENSORMETADATA._serialized_end = 669
    _MODELINFERREQUEST._serialized_start = 672
    _MODELINFERREQUEST._serialized_end = 1550
    _MODELINFERREQUEST_INFERINPUTTENSOR._serialized_start = 980
    _MODELINFERREQUEST_INFERINPUTTENSOR._serialized_end = 1256
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._serialized_start = 1180
    _MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY._serialized_end = 1256
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR._serialized_start = 1259
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR._serialized_end = 1472
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._serialized_start = (
        1180
    )
    _MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY._serialized_end = 1256
    _MODELINFERREQUEST_PARAMETERSENTRY._serialized_start = 1180
    _MODELINFERREQUEST_PARAMETERSENTRY._serialized_end = 1256
    _MODELINFERRESPONSE._serialized_start = 1553
    _MODELINFERRESPONSE._serialized_end = 2150
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR._serialized_start = 1793
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR._serialized_end = 2072
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._serialized_start = 1180
    _MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY._serialized_end = 1256
    _MODELINFERRESPONSE_PARAMETERSENTRY._serialized_start = 1180
    _MODELINFERRESPONSE_PARAMETERSENTRY._serialized_end = 1256
    _INFERPARAMETER._serialized_start = 2152
    _INFERPARAMETER._serialized_end = 2257
    _INFERTENSORCONTENTS._serialized_start = 2260
    _INFERTENSORCONTENTS._serialized_end = 2468
    _REPOSITORYMODELLOADREQUEST._serialized_start = 2470
    _REPOSITORYMODELLOADREQUEST._serialized_end = 2518
    _REPOSITORYMODELLOADRESPONSE._serialized_start = 2520
    _REPOSITORYMODELLOADRESPONSE._serialized_end = 2587
    _REPOSITORYMODELUNLOADREQUEST._serialized_start = 2589
    _REPOSITORYMODELUNLOADREQUEST._serialized_end = 2639
    _REPOSITORYMODELUNLOADRESPONSE._serialized_start = 2641
    _REPOSITORYMODELUNLOADRESPONSE._serialized_end = 2712
    _GRPCINFERENCESERVICE._serialized_start = 2715
    _GRPCINFERENCESERVICE._serialized_end = 3437
# @@protoc_insertion_point(module_scope)



#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsfs import feature_group
from hsfs.engine import python


class TestFeatureGroupWriter:
    def test_fg_writer_context_manager(self, mocker, dataframe_fixture_basic):
        mock_insert = mocker.patch("hsfs.feature_group.FeatureGroup.insert")

        fg = feature_group.FeatureGroup(
            name="test",
            version=1,
            featurestore_id=99,
            primary_key=[],
            partition_key=[],
            id=10,
        )

        with fg.multi_part_insert() as writer:
            assert writer._feature_group == fg
            assert writer._feature_group._multi_part_insert is True
            writer.insert(dataframe_fixture_basic)

        mock_insert.assert_called_once_with(
            features=dataframe_fixture_basic,
            overwrite=False,
            operation="upsert",
            storage=None,
            write_options={"start_offline_materialization": False},
            validation_options={"fetch_expectation_suite": False},
            transformation_context=None,
            transform=True,
        )
        assert fg._multi_part_insert is False

    def test_fg_writer_cache_management(self, mocker, dataframe_fixture_basic):
        engine = python.Engine()
        mocker.patch("hsfs.engine.get_instance", return_value=engine)
        mocker.patch("hopsworks_common.client.get_instance")
        producer, feature_writers, writer_m = (
            mocker.MagicMock(),
            mocker.MagicMock(),
            mocker.MagicMock(),
        )
        headers = {
            "projectId": str(99).encode("utf8"),
            "featureGroupId": str(32).encode("utf8"),
            "subjectId": str(12).encode("utf8"),
        }
        mock_init_kafka_resources = mocker.patch(
            "hsfs.core.kafka_engine._init_kafka_resources",
            return_value=(producer, headers, feature_writers, writer_m),
        )
        mocker.patch("hsfs.core.kafka_engine.encode_complex_features")
        mocker.patch("hsfs.core.job.Job")
        mocker.patch("hsfs.engine.get_type", return_value="python")

        fg = feature_group.FeatureGroup(
            name="test",
            version=1,
            featurestore_id=99,
            primary_key=[],
            partition_key=[],
            id=10,
            features=engine.parse_schema_feature_group(dataframe_fixture_basic, "HUDI"),
            stream=True,
        )
        fg.feature_store = mocker.MagicMock()

        with fg.multi_part_insert() as writer:
            assert writer._feature_group == fg
            assert writer._feature_group._multi_part_insert is True
            writer.insert(dataframe_fixture_basic)

            # after insert cache should be populated
            assert writer._feature_group._multi_part_insert is True
            assert writer._feature_group._kafka_producer == producer
            assert writer._feature_group._feature_writers == feature_writers
            assert writer._feature_group._writer == writer_m
            assert writer._feature_group._kafka_headers == headers

            writer.insert(dataframe_fixture_basic)
            # after second insert should have been called only once
            mock_init_kafka_resources.assert_called_once()

        # leaving context, cache should be reset and call flush
        producer.flush.assert_called_once()
        assert fg._multi_part_insert is False
        assert fg._kafka_producer is None
        assert fg._feature_writers is None
        assert fg._kafka_headers is None
        assert fg._writer is None

    def test_fg_writer_without_context_manager(self, mocker, dataframe_fixture_basic):
        engine = python.Engine()
        mocker.patch("hsfs.engine.get_instance", return_value=engine)
        mocker.patch("hopsworks_common.client.get_instance")
        producer, feature_writers, writer_m = (
            mocker.MagicMock(),
            mocker.MagicMock(),
            mocker.MagicMock(),
        )
        headers = {
            "projectId": str(99).encode("utf8"),
            "featureGroupId": str(32).encode("utf8"),
            "subjectId": str(12).encode("utf8"),
        }
        mock_init_kafka_resources = mocker.patch(
            "hsfs.core.kafka_engine._init_kafka_resources",
            return_value=(producer, headers, feature_writers, writer_m),
        )
        mocker.patch("hsfs.core.kafka_engine.encode_complex_features")
        mocker.patch("hsfs.core.job.Job")
        mocker.patch("hsfs.engine.get_type", return_value="python")

        fg = feature_group.FeatureGroup(
            name="test",
            version=1,
            featurestore_id=99,
            primary_key=[],
            partition_key=[],
            id=10,
            features=engine.parse_schema_feature_group(dataframe_fixture_basic, "HUDI"),
            stream=True,
        )
        fg.feature_store = mocker.MagicMock()

        fg.multi_part_insert(dataframe_fixture_basic)

        # after first insert cache should be populated
        assert fg._multi_part_insert is True
        assert fg._kafka_producer == producer
        assert fg._feature_writers == feature_writers
        assert fg._kafka_headers == headers
        assert fg._writer == writer_m

        fg.multi_part_insert(dataframe_fixture_basic)
        # after second insert should have been called only once
        mock_init_kafka_resources.assert_called_once()

        # finalize should reset cache and call flush
        fg.finalize_multi_part_insert()
        producer.flush.assert_called_once()
        assert fg._multi_part_insert is False
        assert fg._kafka_producer is None
        assert fg._feature_writers is None
        assert fg._kafka_headers is None
        assert fg._writer is None



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json

from hsfs import util


class TrainingDatasetJobConf:
    def __init__(self, query, overwrite, write_options, spark_job_configuration):
        self._query = query
        self._overwrite = overwrite
        self._write_options = write_options
        self._spark_job_configuration = spark_job_configuration

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = query

    @property
    def overwrite(self):
        return self._overwrite

    @overwrite.setter
    def overwrite(self, overwrite):
        self._overwrite = overwrite

    @property
    def write_options(self):
        return self._write_options

    @write_options.setter
    def write_options(self, write_options):
        self._write_options = write_options

    @property
    def spark_job_configuration(self):
        return self._spark_job_configuration

    @spark_job_configuration.setter
    def spark_job_configuration(self, spark_job_configuration):
        self._spark_job_configuration = spark_job_configuration

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "query": self._query,
            "overwrite": self._overwrite,
            "writeOptions": [
                {"name": k, "value": v} for k, v in self._write_options.items()
            ]
            if self._write_options
            else None,
            "sparkJobConfiguration": self._spark_job_configuration,
        }
